# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from decimal import Decimal

import pytest
from pyfakefs.fake_filesystem_unittest import Patcher


@pytest.fixture(autouse=True)
def _mock_fs():
    with Patcher() as patcher:
        patcher.fs.create_dir("/cwd")
        patcher.fs.create_file("/cwd/src/a_file.v")
        patcher.fs.create_file("/cwd/src/another_file.v")
        patcher.fs.create_file("/ncwd/src/a_file.v")
        patcher.fs.create_file("/ncwd/src/another_file.v")
        os.chdir("/cwd")
        yield


def test_expr():
    from .preprocessor import Expr

    assert Expr.evaluate("5 + 4 * (2 + 1)", {}) == 17, "Order of evaluation failure"

    assert (
        Expr.evaluate("5 + 4 * (2 + $A)", {"A": 20}) == 93
    ), "Variable dereferencing failure"

    assert (
        Expr.evaluate("5 + 4 * (2 + $A[0].B)", {"A[0].B": 1}) == 17
    ), "Deep variable dereferencing failure"

    with pytest.raises(TypeError, match="valid numeric"):
        Expr.evaluate("5 + 4 * (2 + $A)", {"A": "20"}) == 93

    with pytest.raises(ValueError, match="is empty"):
        Expr.evaluate("", {})

    with pytest.raises(ValueError, match="multiple values"):
        Expr.evaluate("(5 + 4) (8 + 10)", {})


def test_process_string():
    from .preprocessor import process_string

    assert process_string("expr::2 * 2", {}) == 4, "expr:: not working"

    assert process_string("ref::$A", {"A": "B"}) == "B", "ref:: not working"

    with pytest.raises(KeyError, match="not found"):
        process_string("ref::$A", {})

    assert process_string(
        "refg::$DESIGN_DIR/src/a*.v",
        {"DESIGN_DIR": "/cwd"},
        ["/cwd"],
    ) == [
        "/cwd/src/a_file.v",
        "/cwd/src/another_file.v",
    ], "refg:: in design dir not working"

    assert process_string(
        "refg::$DESIGN_DIR/src/a*.v",
        {"DESIGN_DIR": "/cwd"},
        ["/cwd"],
    ) == process_string(
        "dir::src/a*.v",
        {"DESIGN_DIR": "/cwd"},
        ["/cwd"],
    ), "dir:: doesn't match refg::$DESIGN_DIR"

    with pytest.raises(PermissionError, match="readable to"):
        process_string(
            "refg::$MY_VARIABLE/src/a*.v",
            {"DESIGN_DIR": "/cwd", "MY_VARIABLE": "/ncwd"},
            ["/cwd"],
        )

    assert (
        process_string("refg::$A", {"A": "B"}, ["/cwd"]) == []
    ), "refg:: on non-existent directory not working"


mmpt_raw = {
    "meta": {"version": 2},
    "PDK": "sky130A",
    "STD_CELL_LIBRARY": "sky130_fd_sc_hd",
    "DESIGN_NAME": "manual_macro_placement_test",
    "VERILOG_FILES": "dir::src/*.v",
    "MACROS": {
        "spm": {
            "module": "spm",
            "instances": {
                "spm_inst_0": {"location": [10, 150], "orientation": "N"},
                "spm_inst_1": {
                    "location": [
                        "expr::$MACROS.spm.instances.spm_inst_0.location[1]",
                        150.00,
                    ],
                    "orientation": "N",
                },
            },
        },
    },
}


def test_process_info_extraction():
    from .preprocessor import preprocess_dict

    process_info = preprocess_dict(
        mmpt_raw,
        "/cwd",
        only_extract_process_info=True,
    )

    assert (
        process_info["PDK"] == "sky130A"
    ), "Failed to properly extract PDK info from config"

    assert (
        process_info["STD_CELL_LIBRARY"] == "sky130_fd_sc_hd"
    ), "Failed to properly extract PDK info from config"


def test_preprocess_dict():
    from .preprocessor import preprocess_dict

    preprocessed = preprocess_dict(
        mmpt_raw,
        "/cwd",
        pdk="sky130A",
        pdkpath="/cwd",
        scl="sky130_fd_sc_hd",
        readable_paths=["/cwd"],
    )
    expected = {
        "PDK": "sky130A",
        "PDKPATH": "/cwd",
        "STD_CELL_LIBRARY": "sky130_fd_sc_hd",
        "DESIGN_DIR": "/cwd",
        "meta": {"version": 2},
        "DESIGN_NAME": "manual_macro_placement_test",
        "VERILOG_FILES": ["/cwd/src/a_file.v", "/cwd/src/another_file.v"],
        "MACROS": {
            "spm": {
                "module": "spm",
                "instances": {
                    "spm_inst_0": {"location": [10, 150], "orientation": "N"},
                    "spm_inst_1": {
                        "location": [Decimal("150"), 150.0],
                        "orientation": "N",
                    },
                },
            }
        },
    }
    assert preprocessed == expected, "Preprocessor produced a different result"



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re

from typing import Mapping, Tuple, Dict, Any, Callable, Iterable

modifier_rx = re.compile(r"([\w\-]+)\:([\w\-]+)")


def parse_metric_modifiers(metric_name: str) -> Tuple[str, Mapping[str, str]]:
    """
    Parses a metric name into a base and modifiers as specified in
    the `Metrics4ML standard <https://github.com/ieee-ceda-datc/datc-rdf-Metrics4ML>`_.

    :param metric_name: The name of the metric as generated by a utility.
    :returns: A tuple of the base part as a string, then the modifiers as
        a key-value mapping.
    """
    mn_mut = metric_name.split("__")
    modifiers = {}
    i = len(mn_mut) - 1
    if ":" in mn_mut[i]:
        modifier_list = mn_mut[i].split(":")
        if len(modifier_list) % 2 == 0:
            for i in range(0, len(modifier_list) - 1, 2):
                modifiers[modifier_list[i]] = modifier_list[i + 1]
            mn_mut.pop()
    return "__".join(mn_mut), modifiers


def aggregate_metrics(
    input: Dict[str, Any],
    aggregator_by_metric: Dict[str, Tuple[Any, Callable[[Iterable], Any]]],
) -> Dict[str, Any]:
    aggregated: Dict[str, Any] = {}
    for name, value in input.items():
        metric_name, modifiers = parse_metric_modifiers(name)
        if len(modifiers) != 1:
            # No modifiers = final aggregate, don't double-represent in sums
            # >1 modifiers = n-level nesting, not supported atm
            continue
        entry = aggregator_by_metric.get(metric_name)
        if entry is None:
            continue
        start, aggregator = entry
        current = aggregated.get(metric_name) or start
        aggregated[metric_name] = aggregator([current, value])

    final_values = input.copy()
    final_values.update(aggregated)
    return final_values



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from unittest import mock
from typing import Callable, Optional, Type

import pytest
from pyfakefs.fake_filesystem import FakeFilesystem

from . import flow
from ..config import Variable
from ..config.test_config import (  # noqa: F401
    MOCK_FLOW_VARS,
    mock_variables,
    _mock_fs,
)


@pytest.fixture()
def variable():
    return Variable("DUMMY_VARIABLE", type=str, description="x")


@pytest.fixture()
def MockStepTuple(variable: Variable):
    from ..common import Path
    from ..steps import Step
    from ..state import DesignFormat, State

    class StepA(Step):
        id = "Test.StepA"
        inputs = [DesignFormat.JSON_HEADER]
        outputs = [DesignFormat.JSON_HEADER]

        config_vars = [variable]

        def run(self, state_in: State, **kwargs):
            import json

            out_file = os.path.join(self.step_dir, "whatever.json")
            json.dump(
                {
                    "not_really_a_valid_header": True,
                    "cfg_var_value": self.config["DUMMY_VARIABLE"],
                },
                open(out_file, "w", encoding="utf8"),
            )
            return {
                DesignFormat.JSON_HEADER: Path(out_file),
            }, {"whatever": 0}

    class StepB(Step):
        id = "Test.StepB"
        inputs = []
        outputs = [DesignFormat.JSON_HEADER]

        def run(self, state_in: State, **kwargs):
            import json

            out_file = os.path.join(self.step_dir, "whatever.json")
            json.dump(
                {
                    "probably_a_valid_header": False,
                    "previous_invalid_header": state_in[DesignFormat.JSON_HEADER],
                },
                open(out_file, "w", encoding="utf8"),
            )
            return {
                DesignFormat.JSON_HEADER: Path(out_file),
            }, {"step": 0}

    class StepC(Step):
        id = "Test.StepC"
        inputs = [DesignFormat.JSON_HEADER]
        outputs = [DesignFormat.JSON_HEADER]

        config_vars = [Variable("DUMMY_VARIABLE", type=int, description="x")]

        def run(self, state_in: State, **kwargs):
            import json

            out_file = os.path.join(self.step_dir, "whatever.json")
            json.dump(
                {
                    "not_really_a_valid_header": True,
                    "cfg_var_value": self.config["DUMMY_VARIABLE"],
                },
                open(out_file, "w", encoding="utf8"),
            )
            return {
                DesignFormat.JSON_HEADER: Path(out_file),
            }, {"step": 1}

    return (StepA, StepB, StepC)


@pytest.fixture()
def DummyFlow(MockStepTuple):
    from .flow import Flow

    StepA, StepB, StepC = MockStepTuple

    class Dummy(Flow):
        Steps = [
            StepA,
            StepB,
        ]

        def __init__(
            self,
            *args,
            run_override: Optional[Callable] = None,
            **kwargs,
        ):
            self.run_override = run_override
            super().__init__(*args, **kwargs)

        def run(self, initial_state, **kwargs):
            if self.run_override is not None:
                return self.run_override(self, initial_state, **kwargs)
            return initial_state.copy(), []

    return Dummy


class MockProgress(object):
    def __init__(self, *args, **kwargs):
        self.add_task_called_count = 0
        self.start_called_count = 0
        self.stop_called_count = 0
        self.update_called_count = 0
        self.total = 100
        self.completed = 0
        self.description = "Progress"

    def add_task(self, *args):
        self.add_task_called_count += 1

    def start(self):
        self.start_called_count += 1

    def stop(self):
        self.stop_called_count += 1

    def update(
        self,
        _,
        description: Optional[str] = None,
        total: Optional[int] = None,
        completed: Optional[float] = None,
    ):
        if total_stages := total:
            self.total = total_stages

        if completed_stages := completed:
            self.completed = completed_stages

        if current_task_description := description:
            self.description = current_task_description

        self.update_called_count += 1


# ---


def test_flow_abc_init():
    from . import Flow

    with pytest.raises(TypeError, match="Can't instantiate abstract class") as e:
        Flow()

    assert e is not None, "Flow ABC instantiated successfully"


def test_factory(DummyFlow: Type[flow.Flow]):
    from . import Flow

    assert Flow.factory.list() == [
        "Optimizing",
        "Classic",
        "OpenInKLayout",
        "OpenInOpenROAD",
    ], "One or more built-in flows missing from factory list"

    Flow.factory.register()(DummyFlow)

    assert "Dummy" in Flow.factory.list(), "failed to register new dummy flow"

    Flow.factory.register("AnotherName")(DummyFlow)

    assert (
        "AnotherName" in Flow.factory.list()
    ), "failed to register new dummy flow by another name"

    assert (
        Flow.factory.get("Dummy") == DummyFlow
    ), "failed to retrieve registered dummy flow"


@pytest.mark.usefixtures("_mock_fs")
@mock_variables([flow])
def test_init_and_config_vars(DummyFlow: Type[flow.Flow], variable: Variable):
    flow = DummyFlow(
        {
            "DESIGN_NAME": "WHATEVER",
            "DUMMY_VARIABLE": "PINGAS",
            "VERILOG_FILES": ["/cwd/src/a.v"],
        },
        design_dir="/cwd",
        pdk="dummy",
        scl="dummy_scl",
        pdk_root="/pdk",
    )

    assert flow.get_config_variables() == MOCK_FLOW_VARS + [
        variable
    ], "flow config variables did not match"


@pytest.mark.usefixtures("_mock_fs")
@mock_variables([flow])
def test_clashing_variables(DummyFlow: Type[flow.Flow], MockStepTuple):
    from . import FlowException

    StepA, StepB, StepC = MockStepTuple

    class DummierFlow(DummyFlow):
        Steps = [StepA, StepB, StepC]

    with pytest.raises(
        FlowException,
        match=r"Unrelated variables in \w+ and \w+ share a name",
    ) as e:
        DummierFlow(
            {
                "DESIGN_NAME": "WHATEVER",
                "DUMMY_VARIABLE": "PINGAS",
                "VERILOG_FILES": ["/cwd/src/a.v"],
            },
            design_dir="/cwd",
            pdk="dummy",
            scl="dummy_scl",
            pdk_root="/pdk",
        )

    assert e is not None, "clashing variables did not generate a FlowException"


@pytest.mark.usefixtures("_mock_fs")
@mock_variables([flow])
@mock.patch.object(flow, "Progress", MockProgress)
def test_progress_bar(DummyFlow: Type[flow.Flow]):
    def run_override(self: DummyFlow, initial_state, **kwargs):
        assert (
            self.progress_bar.started
        ), ".start() did not start the progress bar rendering"

        assert (
            self.progress_bar._FlowProgressBar__progress.start_called_count == 1
        ), "start called more than once"

        self.progress_bar.set_max_stage_count(2)
        assert (
            self.progress_bar._FlowProgressBar__progress.total == 2
        ), ".set_max_stage_count() failed to set progress bar total"

        self.progress_bar.start_stage("literally whatever")

        self.progress_bar.end_stage()

        assert (
            self.progress_bar._FlowProgressBar__progress.update_called_count == 3
        ), "unexpected progress bar update count"

        assert (
            self.progress_bar._FlowProgressBar__progress.completed == 1
        ), "task complete count out of sync"

        self.progress_bar.get_ordinal_prefix() == "2-", "incorrect ordinal returned"

        return initial_state.copy(), []

    flow = DummyFlow(
        {
            "DESIGN_NAME": "WHATEVER",
            "DUMMY_VARIABLE": "PINGAS",
            "VERILOG_FILES": ["/cwd/src/a.v"],
        },
        design_dir="/cwd",
        pdk="dummy",
        scl="dummy_scl",
        pdk_root="/pdk",
        run_override=run_override,
    )

    flow.start()


@pytest.mark.usefixtures("_mock_fs")
@mock_variables([flow])
@mock.patch.object(flow, "Progress", MockProgress)
def test_run_tags(
    fs: FakeFilesystem, DummyFlow: Type[flow.Flow], caplog: pytest.LogCaptureFixture
):
    from . import FlowException
    from ..state import State

    flow = DummyFlow(
        {
            "DESIGN_NAME": "WHATEVER",
            "DUMMY_VARIABLE": "PINGAS",
            "VERILOG_FILES": ["/cwd/src/a.v"],
        },
        design_dir="/cwd",
        pdk="dummy",
        scl="dummy_scl",
        pdk_root="/pdk",
    )

    with pytest.raises(
        FlowException, match="tag and last_run cannot be used simultaneously"
    ):
        flow.start(tag="NotNone", last_run=True)

    with pytest.raises(FlowException, match="last_run used without any existing runs"):
        flow.start(last_run=True)

    fs.create_dir("/cwd/runs/MY_TAG")
    flow.start(tag="MY_TAG")
    assert (
        "Starting a new run of the" in caplog.text
    ), ".start() with an empty folder did not print a message about a new run"
    caplog.clear()

    fs.create_dir("/cwd/runs/MY_TAG2")
    fs.create_file(
        "/cwd/runs/MY_TAG2/01-step_a/state_out.json",
        contents=State({}, metrics={"step": 0}).dumps(),
    )
    fs.create_file(
        "/cwd/runs/MY_TAG2/02-step_b/state_out.json",
        contents=State({}, metrics={"step": 1}).dumps(),
    )

    state = flow.start(tag="MY_TAG2")
    assert (
        "Using existing run at" in caplog.text
    ), ".start() with a non-empty folder did not print a message about an existing run"
    assert (
        state.metrics["step"] == 1
    ), ".start() using existing run failed to return latest state"
    caplog.clear()

    flow.start(last_run=True)
    assert flow.run_dir.endswith(
        "MY_TAG2"
    ), ".start() with last_run failed to return latest run"

    fs.create_file("/cwd/runs/MY_TAG3", contents="")
    with pytest.raises(
        FlowException, match="already exists as a file and not a directory"
    ):
        flow.start(tag="MY_TAG3")



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from textwrap import dedent
from functools import partial
from concurrent.futures import ThreadPoolExecutor
from typing import Optional, Union


from click import Parameter, echo
from cloup import (
    option,
    argument,
    option_group,
    Context,
    Choice,
    Path,
)
from cloup.constraints import (
    mutually_exclusive,
)
from cloup.typing import Decorator

from openlane.state.state import InvalidState


from .flow import Flow
from ..common import set_tpe, get_opdks_rev
from ..logging import set_log_level, err, LogLevelsDict
from ..state import State


def set_log_level_cb(
    ctx: Context,
    param: Parameter,
    value: str,
):
    level: Union[str, int] = value
    try:
        try:
            level = int(value)
        except ValueError:
            pass
        set_log_level(level)
    except ValueError as e:
        err(f"Invalid logging level {value}: {e}.")
        echo(ctx.get_help())
        ctx.exit(-1)


def set_worker_count_cb(
    ctx: Context,
    param: Parameter,
    value: Optional[int],
):
    if value is None:
        return None

    set_tpe(ThreadPoolExecutor(max_workers=value))


def initial_state_cb(
    ctx: Context,
    param: Parameter,
    value: Optional[str],
):
    if value is None:
        return None

    try:
        initial_state_str = open(value, encoding="utf8").read()
    except Exception as e:
        err(f"Failed to read initial state: {e}")
        ctx.exit(-1)
    try:
        initial_state = State.loads(initial_state_str, validate_path=True)
    except InvalidState as e:
        err(e)
        ctx.exit(-1)

    return initial_state


def only_cb(
    ctx: Context,
    param: Parameter,
    value: Optional[str],
):
    if value is not None:
        ctx.obj = ctx.obj or {}
        ctx.obj["only"] = value
    return value


def from_to_cb(
    ctx: Context,
    param: Parameter,
    value: Optional[str],
):
    if isinstance(ctx.obj, dict) and ctx.obj.get("only"):
        return ctx.obj.get("only")
    return value


def use_volare_cb(
    ctx: Context,
    param: Parameter,
    value: bool,
):
    ctx.obj = ctx.obj or {}
    ctx.obj["use_volare"] = value


def pdk_scl_cb(
    ctx: Context,
    param: Parameter,
    value: Optional[str],
):
    if param.name is None:
        return

    values = ctx.params.copy()
    values[param.name] = value
    if "pdk" in values and "scl" in values:
        pdk = values["pdk"]
        pdk_family = pdk[:-1]
        if ctx.obj and ctx.obj.get("use_volare"):
            import volare

            pdk_root = volare.get_volare_home(values["pdk_root"])

            volare.enable(pdk_root, pdk_family, get_opdks_rev())
    return value


def cloup_flow_opts(
    config_options: bool = True,
    run_options: bool = True,
    sequential_flow_controls: bool = True,
    pdk_options: bool = True,
    log_level: bool = True,
    jobs: bool = True,
    accept_config_files: bool = True,
    volare_by_default: bool = True,
) -> Decorator:
    """
    Creates a wrapper that appends a number of OpenLane flow-related flags to a
    function decorated with `@cloup.command https://cloup.readthedocs.io/en/stable/autoapi/cloup/index.html#cloup.command`_.

    The following keyword arguments will be passed to the decorated function.
    * Those postfixed ‡ are compatible with the constructor for :class:`Flow`.
    * Those postfixed § are compatible with the :meth:`Flow.start`.

    * Flow configuration options (if parameter ``config_options`` is ``True``):
        * ``flow_name``: ``Optional[str]``: A valid flow ID to be used with :meth:`Flow.factory.get`
        * ``config_override_strings``‡: ``Optional[Iterable[str]]``
    * Sequential flow controls (if parameter ``sequential_flow_controls`` is ``True``)
        * ``frm``§: ``str``: Start from a step with this ID. Supported by sequential flows.
        * ``to``§: ``str``: Stop at a step with this id. Supported by sequential flows.
        * ``skip``§: ``Iterable[str]``: Skip these steps. Supported by sequential flows.
    * Flow run options (if parameter ``run_options`` is ``True``):
        * ``tag``§: ``Optional[str]``
        * ``last_run``§: ``bool``: If ``True``, ``tag`` is guaranteed to be None.
        * ``with_initial_state``§: ``Optional[State]``
    * PDK options
        * ``use_volare``: ``bool``
        * ``pdk_root``‡: ``Optional[str]``
        * ``pdk``‡: ``str``
        * ``scl``‡: ``Optional[str]``
    * ``config_files``: ``Iterable[str]``: Paths to configuration files (if
      parameter  ``accept_config_files`` is ``True``)

    :param config_options: Enables flow configuration and starting CLI flags
    :param sequential_flow_controls: Enables flow control CLI flags
    :param flow_run_options: Enables tag CLI flags
    :param pdk_options: Enables PDK CLI flags
    :param log_level: Enables ``--log-level`` CLI flag
    :param jobs: Enables ``-j/--jobs`` CLI flag
    :param accept_config_files: Accepts configuration file paths as CLI arguments
    :param volare_by_default: If ``pdk_options`` is ``True``, this changes whether
        Volare is used by default for this CLI or not.
    :returns: The wrapper
    """
    o = partial(option, show_default=True)

    def decorator(f):
        if config_options:
            f = option_group(
                "Flow configuration options",
                o(
                    "-f",
                    "--flow",
                    "flow_name",
                    type=Choice(Flow.factory.list(), case_sensitive=False),
                    default=None,
                    help="The built-in OpenLane flow to use for this run",
                ),
                o(
                    "-c",
                    "--override-config",
                    "config_override_strings",
                    type=str,
                    multiple=True,
                    help="For this run only- override a configuration variable with a certain value. In the format KEY=VALUE. Can be specified multiple times. Values must be valid JSON values.",
                ),
            )(f)
        if run_options:
            f = option_group(
                "Run options",
                o(
                    "--run-tag",
                    "tag",
                    default=None,
                    type=str,
                    help="An optional name to use for this particular run of an OpenLane-based flow. Used to create the run directory.",
                ),
                o(
                    "-i",
                    "--with-initial-state",
                    type=Path(
                        exists=True,
                        file_okay=True,
                        dir_okay=False,
                    ),
                    default=None,
                    callback=initial_state_cb,
                    help="Use this JSON file as an initial state. If this is not specified, the latest `state_out.json` of the run directory will be used if available.",
                ),
                o(
                    "--last-run",
                    is_flag=True,
                    default=False,
                    help="Use the last run as the run tag.",
                ),
                constraint=mutually_exclusive,
            )(f)
        if sequential_flow_controls:
            f = option_group(
                "Sequential flow controls",
                o(
                    "-F",
                    "--from",
                    "frm",
                    type=str,
                    default=None,
                    callback=from_to_cb,
                    help="Start from a step with this id. Supported by sequential flows.",
                ),
                o(
                    "-T",
                    "--to",
                    type=str,
                    default=None,
                    callback=from_to_cb,
                    help="Stop at a step with this id. Supported by sequential flows.",
                ),
                o(
                    "--only",
                    type=str,
                    default=None,
                    expose_value=False,
                    is_eager=True,
                    callback=only_cb,
                    help="Shorthand to set both --from and --to to the same value. Overrides the values from both.",
                ),
                o(
                    "-S",
                    "--skip",
                    type=str,
                    multiple=True,
                    help="Skip these steps. Supported by sequential flows.",
                ),
            )(f)
        if log_level:
            f = o(
                "--log-level",
                type=str,
                default="VERBOSE",
                help=dedent(
                    """
                    A logging level. Set to INFO or higher to silence subprocess logs.

                    You can provide either a number or a string out of the following (higher is more silent):
                    """
                )
                + ",".join(
                    [f"{name}={value}" for name, value in LogLevelsDict.items()]
                ),
                callback=set_log_level_cb,
                expose_value=False,
            )(f)
        if pdk_options:
            f = option_group(
                "PDK options",
                o(
                    "--volare-pdk/--manual-pdk",
                    "use_volare",
                    is_eager=True,
                    default=volare_by_default,
                    help="Automatically use Volare for PDK version installation and enablement. Set --manual if you want to use a custom PDK version.",
                    expose_value=False,
                    callback=use_volare_cb,
                ),
                o(
                    "--pdk-root",
                    type=Path(
                        exists=True,
                        file_okay=False,
                        dir_okay=True,
                    ),
                    is_eager=True,
                    default=os.environ.pop("PDK_ROOT", None),
                    help="Override volare PDK root folder. Required if Volare is not installed.",
                ),
                o(
                    "-p",
                    "--pdk",
                    type=str,
                    default=os.environ.pop("PDK", "sky130A"),
                    help="The process design kit to use.",
                    callback=pdk_scl_cb,
                ),
                o(
                    "-s",
                    "--scl",
                    type=str,
                    default=os.environ.pop("STD_CELL_LIBRARY", None),
                    help="The standard cell library to use. If None, the PDK's default standard cell library is used.",
                    callback=pdk_scl_cb,
                ),
            )(f)
        if jobs:
            f = o(
                "-j",
                "--jobs",
                type=int,
                default=os.cpu_count(),
                help="The maximum number of threads or processes that can be used by OpenLane.",
                callback=set_worker_count_cb,
                expose_value=False,
            )(f)
        if accept_config_files:
            f = argument(
                "config_files",
                nargs=-1,
                type=Path(
                    exists=True,
                    file_okay=True,
                    dir_okay=True,
                ),
            )(f)
        return f

    return decorator



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import os
from typing import Iterable, List, Set, Tuple, Optional, Type, Dict, Union

from .flow import Flow, FlowException, FlowError
from ..state import State
from ..steps import (
    Step,
    StepError,
    StepException,
    DeferredStepError,
)
from ..logging import info, success, err


class SequentialFlow(Flow):
    """
    The simplest Flow, running each Step as a stage, serially,
    with nothing happening in parallel and no significant inter-step
    processing.

    All subclasses of this flow have to do is override the :attr:`.Steps` abstract property
    and it would automatically handle the rest. See `Classic` in Built-in Flows for an example.

    It should be noted, for Steps with duplicate IDs, all Steps other than the
    first one will technically be subclassed with no change other than to simply
    set the ID to the previous step's ID with a suffix: i.e. the second instance
    of ``Test.MyStep`` will have an ID of ``Test.MyStep1``, and so on.

    :param Substitute: Substitute all instances of one `Step` type by another `Step`
        type in the :attr:`.Steps` attribute for this instance only.

        You may also use the string Step IDs in place of a `Step` type object.

        Duplicate ID normalization is re-run after substitutions.

    :param args: Arguments for :class:`Flow`.
    :param kwargs: Keyword arguments for :class:`Flow`.
    """

    def __init_subclass__(cls, scm_type=None, name=None, **kwargs):
        cls.__normalize_step_ids(cls)

    @classmethod
    def make(Self, step_ids: List[str]) -> Type[SequentialFlow]:
        Step_list = []
        for name in step_ids:
            step = Step.factory.get(name)
            if step is None:
                raise TypeError(f"No step found with id '{name}'")
            Step_list.append(step)

        class CustomSequentialFlow(SequentialFlow):
            name = "Custom Sequential Flow"
            Steps = Step_list

        return CustomSequentialFlow

    @staticmethod
    def __normalize_step_ids(target: Union[SequentialFlow, Type[SequentialFlow]]):
        ids_used: Set[str] = set()

        for i, step in enumerate(target.Steps):
            counter = 0
            id = step.id
            name = step.__name__
            while id in ids_used:
                counter += 1
                id = f"{step.id}-{counter}"
                name = f"{step.__name__}_{counter}"
            if id != step.id:
                new_step = type(name, (step,), {"id": id})
                target.Steps[i] = new_step
            ids_used.add(id)

    def __init__(
        self,
        *args,
        Substitute: Optional[
            Dict[str, Union[str, Type[Step]]],
        ] = None,
        **kwargs,
    ):
        self.Steps = self.Steps.copy()  # Break global reference

        if substitute := Substitute:
            for key, item in substitute.items():
                self.__substitute_step(key, item)

            self.__normalize_step_ids(self)

        super().__init__(*args, **kwargs)

    def __substitute_step(
        self,
        id: str,
        with_step: Union[str, Type[Step]],
    ):
        step_indices: List[int] = []
        for i, step in enumerate(self.Steps):
            if step.id.lower() == id.lower():
                step_indices.append(i)

        if len(step_indices) == 0:
            raise FlowException(
                f"Could not substitute '{id}' with '{with_step}': no steps with ID '{id}' found in flow."
            )

        if isinstance(with_step, str):
            with_step_opt = Step.factory.get(with_step)
            if with_step_opt is None:
                raise FlowException(
                    f"Could not substitute '{step.id}' with '{with_step}': no replacement step with ID '{with_step}' found."
                )
            with_step = with_step_opt

        for i in step_indices:
            self.Steps[i] = with_step

    def run(
        self,
        initial_state: State,
        frm: Optional[str] = None,
        to: Optional[str] = None,
        skip: Optional[Iterable[str]] = None,
        **kwargs,
    ) -> Tuple[State, List[Step]]:
        step_ids = {cls.id.lower(): cls.id for cls in reversed(self.Steps)}
        skipped_ids = []

        frm_resolved = None
        if frm is not None:
            if id := step_ids.get(frm.lower()):
                frm_resolved = id
            else:
                raise FlowException(
                    f"Failed to process start step '{frm}': no step with ID '{frm}' found in flow."
                )

        to_resolved = None
        if to is not None:
            if id := step_ids.get(to.lower()):
                to_resolved = id
            else:
                raise FlowException(
                    f"Failed to process start step '{to}': no step with ID '{to}' found in flow."
                )

        if skipped_steps := skip:
            for skipped_step in skipped_steps:
                if id := step_ids.get(skipped_step.lower()):
                    skipped_ids.append(id)
                else:
                    raise FlowException(
                        f"Failed to process skipped step '{skipped_step}': no step with ID '{skipped_step}' found in flow."
                    )
        step_count = len(self.Steps)
        self.progress_bar.set_max_stage_count(step_count)

        step_list = []

        info("Starting…")

        executing = frm is None
        deferred_errors = []

        current_state = initial_state
        for cls in self.Steps:
            step = cls(config=self.config, state_in=current_state)
            if frm_resolved is not None and frm_resolved == step.id:
                executing = True

            self.progress_bar.start_stage(step.name)
            increment_ordinal = True
            if not executing or cls.id in skipped_ids:
                info(f"Skipping step '{step.name}'…")
                increment_ordinal = False
            else:
                step_list.append(step)
                try:
                    current_state = step.start(
                        toolbox=self.toolbox,
                        step_dir=self.dir_for_step(step),
                    )
                except StepException as e:
                    raise FlowException(str(e))
                except DeferredStepError as e:
                    deferred_errors.append(str(e))
                except StepError as e:
                    raise FlowError(str(e))

            self.progress_bar.end_stage(increment_ordinal=increment_ordinal)

            if to_resolved and to_resolved == step.id:
                executing = False
        if len(deferred_errors) != 0:
            err("The following deferred step errors have been encountered:")
            for error in deferred_errors:
                err(error)
            raise FlowError("One or more deferred errors were encountered.")

        assert self.run_dir is not None
        final_views_path = os.path.join(self.run_dir, "final")
        info(f"Saving final views to '{final_views_path}'…")
        try:
            current_state.save_snapshot(final_views_path)
        except Exception as e:
            raise FlowException(f"Failed to save final views: {e}")
        success("Flow complete.")
        return (current_state, step_list)



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
import click
import atexit
import logging
import rich.console
from typing import ClassVar, Union
from rich.logging import RichHandler


class LogLevels:
    ALL: ClassVar[int] = 0
    DEBUG: ClassVar[int] = 10
    VERBOSE: ClassVar[int] = 15
    INFO: ClassVar[int] = 20
    WARN: ClassVar[int] = 30
    ERROR: ClassVar[int] = 40
    CRITICAL: ClassVar[int] = 50


def __log_levels_dict():
    result = {k: v for k, v in LogLevels.__dict__.items() if not k.startswith("__")}
    for name, value in result.items():
        logging.addLevelName(value, name)
    return result


LogLevelsDict = __log_levels_dict()

console = rich.console.Console()
atexit.register(lambda: rich.console.Console().show_cursor())
__plain_output = "pytest" in sys.modules


class NullFormatter(logging.Formatter):
    def format(self, record):
        return record.getMessage()


def __logger():
    handler: logging.Handler
    if __plain_output:
        handler = logging.StreamHandler()
        handler.setFormatter(NullFormatter())
    else:
        handler = RichHandler(
            console=console,
            rich_tracebacks=True,
            markup=True,
            tracebacks_suppress=[
                click,
            ],
            show_level=False,
        )
        handler.setFormatter(logging.Formatter("%(message)s", datefmt="[%X]"))
    logger = logging.getLogger("__openlane__")
    logger.setLevel(LogLevels.VERBOSE)
    logger.addHandler(handler)
    return logger


__openlane_logger = __logger()


def set_log_level(lv: Union[str, int]):
    __openlane_logger.setLevel(lv)


def reset_log_level():
    set_log_level("VERBOSE")


def get_log_level() -> int:
    return __openlane_logger.getEffectiveLevel()


def debug(msg: object, /, **kwargs):
    """
    Logs to the OpenLane logger with the log level DEBUG.

    :param msg: The message to log
    """
    if kwargs.get("stacklevel") is None:
        kwargs["stacklevel"] = 2
    __openlane_logger.debug(msg, **kwargs)


def verbose(*args, **kwargs):
    """
    Prints to the console if the log level is <= VERBOSE.

    All args and kwargs are passed to https://rich.readthedocs.io/en/stable/reference/console.html#rich.console.Console.print
    """
    if get_log_level() > LogLevels.VERBOSE:
        return
    if __plain_output:
        print(*args, **kwargs)
    else:
        console.print(*args, **kwargs)


def info(msg: object, /, **kwargs):
    """
    Logs to the OpenLane logger with the log level INFO.

    :param msg: The message to log
    """
    if kwargs.get("stacklevel") is None:
        kwargs["stacklevel"] = 2
    __openlane_logger.info(msg, **kwargs)


def rule(title: str = "", /, **kwargs):  # pragma: no cover
    """
    Prints a horizontal line on the terminal enclosing the first argument
    if the log level is <= INFO.

    Kwargs are passed to https://rich.readthedocs.io/en/stable/reference/console.html#rich.console.Console.rule

    :param title: A title string to enclose in the console rule
    """
    if __plain_output:
        print(("-" * 10) + str(title) + ("-" * 10))
    else:
        console.rule(title, **kwargs)


def success(msg: object, /, **kwargs):
    """
    Logs an item to the OpenLane logger with a success unicode character and
    green/bold rich formatting syntax with the log level INFO.

    :param msg: The message to log
    """
    if kwargs.get("stacklevel") is None:
        kwargs["stacklevel"] = 2
    __openlane_logger.info(f"⭕ [green][bold] {msg}", **kwargs)


def warn(msg: object, /, **kwargs):
    """
    Logs an item to the OpenLane logger with a warning unicode character and
    gold/bold rich formatting syntax with the log level WARN.

    :param msg: The message to log
    """
    if kwargs.get("stacklevel") is None:
        kwargs["stacklevel"] = 2
    __openlane_logger.warning(f"⚠️  [gold][bold] {msg}", **kwargs)


def err(msg: object, /, **kwargs):
    """
    Logs an item to the OpenLane logger terminal with an error unicode character and
    red/bold rich formatting syntax with the log level ERROR.
    """
    if kwargs.get("stacklevel") is None:
        kwargs["stacklevel"] = 2
    __openlane_logger.error(f"❌ [red][bold] {msg}", **kwargs)


if __name__ == "__main__":
    debug("Debug")
    verbose("Verbose")
    rule()
    info("Info")
    success("Success")
    warn("Warn")
    err("Err")



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Type

import pytest

from . import flow
from ..steps import Step
from ..config.test_config import (  # noqa: F401
    mock_variables,
    _mock_fs,
)


@pytest.fixture()
def MetricIncrementer():
    from ..steps import Step

    @Step.factory.register()
    class MetricIncrementer(Step):
        id = "Test.MetricIncrementer"
        inputs = []
        outputs = []

        counter_name = "counter"

        def run(self, state_in, **kwargs):
            metric_to_increment = state_in.metrics.get(self.counter_name, 0)
            metric_to_increment += 1
            return {}, {self.counter_name: metric_to_increment}

    return MetricIncrementer


@pytest.mark.usefixtures("_mock_fs")
@mock_variables([flow])
def test_sequential_flow(MetricIncrementer: Type[Step]):
    from . import SequentialFlow

    class Dummy(SequentialFlow):
        Steps = [
            MetricIncrementer,
            MetricIncrementer,
            MetricIncrementer,
            MetricIncrementer,
        ]

    flow = Dummy(
        {
            "DESIGN_NAME": "WHATEVER",
            "VERILOG_FILES": ["/cwd/src/a.v"],
        },
        design_dir="/cwd",
        pdk="dummy",
        scl="dummy_scl",
        pdk_root="/pdk",
    )

    assert [step.id for step in flow.Steps] == [
        "Test.MetricIncrementer",
        "Test.MetricIncrementer-1",
        "Test.MetricIncrementer-2",
        "Test.MetricIncrementer-3",
    ], "SequentialFlow did not increment IDs properly for duplicate steps"

    state = flow.start()
    assert state.metrics["counter"] == 4, "SequentialFlow did not run properly"


@pytest.mark.usefixtures("_mock_fs")
@mock_variables([flow])
def test_custom_seqflow(MetricIncrementer):
    from . import SequentialFlow

    MyFlow = SequentialFlow.make(
        [
            "Test.MetricIncrementer",
            "Test.MetricIncrementer",
            "Test.MetricIncrementer",
            "Test.MetricIncrementer",
        ]
    )

    for step in MyFlow.Steps:
        assert issubclass(
            step, MetricIncrementer
        ), "CustomSequentialFlow could not properly import Steps by id"

    flow = MyFlow(
        {
            "DESIGN_NAME": "WHATEVER",
            "VERILOG_FILES": ["/cwd/src/a.v"],
        },
        design_dir="/cwd",
        pdk="dummy",
        scl="dummy_scl",
        pdk_root="/pdk",
    )

    state = flow.start()
    assert state.metrics["counter"] == 4, "CustomSequentialFlow did not run properly"


@pytest.mark.usefixtures("_mock_fs")
@mock_variables([flow])
def test_custom_seqflow_bad_id(MetricIncrementer):
    from . import SequentialFlow

    with pytest.raises(TypeError, match="No step found with id"):
        SequentialFlow.make(
            [
                "Test.MetricIncrementer",
                "Test.MetricIncrementer",
                "Test.NotARealIncrement",
                "Test.MetricIncrementer",
            ]
        )


@pytest.mark.usefixtures("_mock_fs")
@mock_variables([flow])
def test_substitution(MetricIncrementer):
    from . import SequentialFlow

    @Step.factory.register()
    class OtherMetricIncrementer(MetricIncrementer):
        id = "Test.OtherMetricIncrementer"
        counter_name = "other_counter"

    MyFlow = SequentialFlow.make(
        [
            "Test.MetricIncrementer",
            "Test.MetricIncrementer",
            "Test.MetricIncrementer",
            "Test.MetricIncrementer",
        ]
    )

    flow = MyFlow(
        {
            "DESIGN_NAME": "WHATEVER",
            "VERILOG_FILES": ["/cwd/src/a.v"],
        },
        design_dir="/cwd",
        pdk="dummy",
        scl="dummy_scl",
        pdk_root="/pdk",
        Substitute={
            "Test.MetricIncrementer": OtherMetricIncrementer,
            "Test.MetricIncrementer-1": "Test.OtherMetricIncrementer",
        },
    )

    assert [step.id for step in flow.Steps] == [
        "Test.OtherMetricIncrementer",
        "Test.OtherMetricIncrementer-1",
        "Test.MetricIncrementer-2",
        "Test.MetricIncrementer-3",
    ], "SequentialFlow did not increment IDs properly for duplicate steps"

    state = flow.start()
    assert (
        state.metrics["other_counter"] == 2
    ), "step substitution replaced other than exactly two step"


@pytest.mark.usefixtures("_mock_fs")
@mock_variables([flow])
def test_bad_substitution(MetricIncrementer):
    from . import SequentialFlow, FlowException

    MyFlow = SequentialFlow.make(
        [
            "Test.MetricIncrementer",
            "Test.MetricIncrementer",
            "Test.MetricIncrementer",
            "Test.MetricIncrementer",
        ]
    )

    with pytest.raises(
        FlowException, match=r"no replacement step with ID '[\w\.]+' found"
    ):
        MyFlow(
            {
                "DESIGN_NAME": "WHATEVER",
                "VERILOG_FILES": ["/cwd/src/a.v"],
            },
            design_dir="/cwd",
            pdk="dummy",
            scl="dummy_scl",
            pdk_root="/pdk",
            Substitute={
                "Test.MetricIncrementer": "Test.NotARealStep",
            },
        )
    with pytest.raises(
        FlowException, match=r"no steps with ID '[\w\.]+' found in flow"
    ):
        MyFlow(
            {
                "DESIGN_NAME": "WHATEVER",
                "VERILOG_FILES": ["/cwd/src/a.v"],
            },
            design_dir="/cwd",
            pdk="dummy",
            scl="dummy_scl",
            pdk_root="/pdk",
            Substitute={
                "Test.NotAStepInTheFlow": "Test.MetricIncrementer",
            },
        )


@pytest.mark.usefixtures("_mock_fs")
@mock_variables([flow])
def test_flow_control(MetricIncrementer):
    from . import SequentialFlow

    class OtherMetricIncrementer(MetricIncrementer):
        id = "Test.OtherMetricIncrementer"
        counter_name = "other_counter"

    class AnotherMetricIncrementer(MetricIncrementer):
        id = "Test.AnotherMetricIncrementer"
        counter_name = "another_counter"

    class YetAnotherMetricIncrementer(MetricIncrementer):
        id = "Test.YetAnotherMetricIncrementer"
        counter_name = "yet_another_counter"

    class LastMetricIncrementer(MetricIncrementer):
        id = "Test.LastMetricIncrementer"
        counter_name = "last_another_counter"

    class Dummy(SequentialFlow):
        Steps = [
            MetricIncrementer,
            OtherMetricIncrementer,
            AnotherMetricIncrementer,
            YetAnotherMetricIncrementer,
            LastMetricIncrementer,
        ]

    flow = Dummy(
        {
            "DESIGN_NAME": "WHATEVER",
            "VERILOG_FILES": ["/cwd/src/a.v"],
        },
        design_dir="/cwd",
        pdk="dummy",
        scl="dummy_scl",
        pdk_root="/pdk",
    )
    state = flow.start(
        frm="test.othermetricincrementer",
        to="test.yetanothermetricincrementer",
        skip=["test.anothermetricincrementer"],
    )
    assert list(state.metrics.keys()) == [
        "other_counter",
        "yet_another_counter",
    ], "flow control did not yield the expected results"



import os

import pytest

from ..common import Path
from pyfakefs.fake_filesystem_unittest import Patcher
from typing import Dict


@pytest.fixture()
def _mock_fs():
    with Patcher() as patcher:
        patcher.fs.create_dir("/cwd")
        os.chdir("/cwd")
        yield


def test_create_by_type():
    from .state import State
    from .design_format import DesignFormat

    test_dict = {DesignFormat.NETLIST: "abc"}
    state = State(test_dict)
    assert state[DesignFormat.NETLIST] == "abc"


def test_create_by_id():
    from .state import State
    from .design_format import DesignFormat

    test_dict = {"nl": "abc"}
    state = State(test_dict)
    assert state[DesignFormat.NETLIST] == "abc"


def test_override_by_id():
    from .state import State
    from .design_format import DesignFormat

    test_dict = {"nl": "abc"}
    override = {"nl": "abcd"}
    state = State(test_dict, overrides=override)
    assert state[DesignFormat.NETLIST] == "abcd"


def test_override_by_type():
    from .state import State
    from .design_format import DesignFormat

    test_dict = {"nl": "abc"}
    override = {DesignFormat.NETLIST: "abcd"}
    state = State(test_dict, overrides=override)
    assert state[DesignFormat.NETLIST] == "abcd"


def test_immutable():
    from .state import State
    from .design_format import DesignFormat

    test_dict = {DesignFormat.NETLIST: "abc"}
    state = State(test_dict)
    with pytest.raises(TypeError, match="State is immutable"):
        state[DesignFormat.NETLIST] = "abcd"

    with pytest.raises(TypeError, match="State is immutable"):
        del state[DesignFormat.NETLIST]


def test_to_raw_dict():
    from .state import State
    from .design_format import DesignFormat

    test_dict = {DesignFormat.NETLIST: "abc"}
    test_metrics = {"metric": "a"}
    state = State(test_dict, metrics=test_metrics)
    raw_dict = state.to_raw_dict()

    assert isinstance(raw_dict, Dict)
    assert raw_dict["nl"] == "abc"
    assert raw_dict["metrics"]["metric"] == "a"


def test_metrics_access():
    from .state import State

    test_dict = {}
    test_metrics = {"metric": "a"}
    state = State(test_dict, metrics=test_metrics)
    assert state.metrics["metric"] == "a"


def test_metrics_mutate():
    from .state import State

    test_dict = {}
    test_metrics = {"metric": "a"}
    state = State(test_dict, metrics=test_metrics)
    with pytest.raises(TypeError, match="is immutable"):
        state.metrics["metric"] = "c"

    with pytest.raises(TypeError, match="is immutable"):
        del state.metrics["metric"]

    with pytest.raises(TypeError, match="is immutable"):
        del state.metrics

    with pytest.raises(TypeError, match="is immutable"):
        state.metrics = {"metric": "b"}


def test_copy():
    from .state import State
    from .design_format import DesignFormat

    test_dict = {"nl": "abc", "spef": {"nom": "abc"}}
    test_metrics = {"metric": "a"}
    state = State(test_dict, metrics=test_metrics)
    state_copy = state.copy()

    assert state_copy[DesignFormat.NETLIST] == "abc"
    assert state_copy.metrics == test_metrics


def test_empty():
    from .state import State
    from .design_format import DesignFormat

    test_dict = {}
    test_metrics = {}
    state = State(test_dict, metrics=test_metrics)

    assert state.metrics == test_metrics
    for format in DesignFormat:
        assert state[format.value.id] is None, "new state has non-none value"


def test_path_fail_exists():
    from .state import State, InvalidState

    test_dict = {"nl": "./state.py", "pnl": "abcd"}
    test_metrics = {"metric": "a"}
    state = State(test_dict, metrics=test_metrics)

    with pytest.raises(InvalidState, match="is not a Path"):
        state.validate()


# def test_path_fail_extension():
#    from .state import State, InvalidState
#
#    test_dict = {"nl": Path("./state.py")}
#    test_metrics = {"metric": "a"}
#    state = State(test_dict, metrics=test_metrics)
#
#    with pytest.raises(InvalidState, match="is not a Path"):
#        state.validate()


@pytest.mark.usefixtures("_mock_fs")
def test_path_success():
    from .state import State

    test_file = "test.nl.v"
    with open(test_file, "w") as f:
        f.write("\n")

    test_dict = {"nl": Path(test_file)}
    test_metrics = {"metric": "a"}
    state = State(test_dict, metrics=test_metrics)

    state.validate()


@pytest.mark.usefixtures("_mock_fs")
def test_save():
    import json
    from .state import State

    test_file = "test.nl.v"
    test_file_contents = "test\n"
    with open(test_file, "w") as f:
        f.write(test_file_contents)

    test_dict = {"spef": {"nom": Path(test_file)}, "nl": Path(test_file)}
    test_metrics = {
        "metric": "a",
        "metric1": 1,
        "metric2": True,
        "metric3": Path("path"),
    }
    state = State(test_dict, metrics=test_metrics)

    save_path = "out"

    state.save_snapshot(save_path)

    save_metrics_path = os.path.join(save_path, "metrics.json")
    save_nl_path = os.path.join(save_path, "nl", test_file)
    save_spef_path = os.path.join(save_path, "spef", "nom", test_file)
    assert os.path.exists(save_path), "save_snapshot failed to create directory"
    assert os.path.exists(save_metrics_path), "save_snapshot failed to create metrics"
    assert os.path.exists(save_nl_path), "save_snapshot failed to create netlist"
    assert os.path.exists(save_spef_path), "save_snapshot failed to create spef"

    assert (
        json.load(open(save_metrics_path, encoding="utf8")) == test_metrics
    ), "serialized metrics does not match previous object"

    f = open(save_nl_path, encoding="utf8")
    save_nl_contents = f.read()
    f.close()
    assert save_nl_contents == test_file_contents

    f = open(save_spef_path, encoding="utf8")
    save_spef_contents = f.read()
    f.close()
    assert save_spef_contents == test_file_contents


@pytest.mark.usefixtures("_mock_fs")
def test_loads():
    import json
    from .state import State

    test_file = "test.nl.v"
    test_file_contents = "test\n"
    with open(test_file, "w") as f:
        f.write(test_file_contents)

    test_dict = {"nl": test_file}
    test_metrics = {"metric": "a", "metric1": 1, "metric2": True, "metric3": "path"}
    state = State(test_dict, metrics=test_metrics)

    new_state = State.loads(json.dumps(state.to_raw_dict()))
    assert new_state.to_raw_dict() == state.to_raw_dict()



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import io
from decimal import Decimal
import pytest

MAGIC_EXAMPLE = """RAM8
----------------------------------------
P-diff distance to N-tap must be < 15.0um (LU.3)
----------------------------------------
17.990um 21.995um 18.265um 22.995um
20.905um 22.935um 21.575um 22.995um
18.535um 21.995um 18.795um 22.635um
"""


def test_magic_drc():

    from . import DRC, Violation

    drc_object, count = DRC.from_magic(io.StringIO(MAGIC_EXAMPLE))
    violations = {
        "LU.3": Violation(
            rules=[("LU", "3")],
            description="P-diff distance to N-tap must be < 15.0um (LU.3)",
            bounding_boxes=[
                (
                    Decimal("17.990"),
                    Decimal("21.995"),
                    Decimal("18.265"),
                    Decimal("22.995"),
                ),
                (
                    Decimal("20.905"),
                    Decimal("22.935"),
                    Decimal("21.575"),
                    Decimal("22.995"),
                ),
                (
                    Decimal("18.535"),
                    Decimal("21.995"),
                    Decimal("18.795"),
                    Decimal("22.635"),
                ),
            ],
        )
    }

    assert drc_object.module == "RAM8", "Failed to extract module object"
    assert count == 3, "Incorrect module count"
    assert violations == drc_object.violations, "Violation mismatch"


def test_magic_drc_badrule():
    from . import DRC

    description = "P-diff distance to N-tap must be < 15.0um (egg salad)"
    magic_bad_rule_example = f"""RAM8
    ----------------------------------------
    {description}
    ----------------------------------------
    17.990um 21.995um 18.265um 22.995um
    20.905um 22.935um 21.575um 22.995um
    18.535um 21.995um 18.795um 22.635um
    """

    drc_object, _ = DRC.from_magic(io.StringIO(magic_bad_rule_example))
    assert (
        drc_object.violations["UNKNOWN.UNKNOWN0"].description == description
    ), "bad rule name improperly handled"


def test_magic_drc_exceptions():
    from . import DRC

    BAD_MAGIC_EXAMPLE = """
    ----------------------------------------
    P-diff distance to N-tap must be < 15.0um (LU.3)
    ----------------------------------------
    invalid example bounding box
    """

    with pytest.raises(ValueError, match="number is invalid"):
        DRC.from_magic(io.StringIO(BAD_MAGIC_EXAMPLE))

    BAD_MAGIC_EXAMPLE_2 = """
    ----------------------------------------
    P-diff distance to N-tap must be < 15.0um (LU.3)
    ----------------------------------------
    18.535um 21.995um 18.795um 22.635um 22.635um
    """

    with pytest.raises(ValueError, match="bounding box has 5/4 elements"):
        DRC.from_magic(io.StringIO(BAD_MAGIC_EXAMPLE_2))


def test_klayout_xml():
    from . import DRC
    from xml.etree import ElementTree as ET

    drc_object, _ = DRC.from_magic(io.StringIO(MAGIC_EXAMPLE))

    bio = io.BytesIO()
    drc_object.to_klayout_xml(bio)

    try:
        parsed = ET.fromstring(bio.getvalue().decode("utf8"))
    except Exception as e:
        pytest.fail(f"Unexpected error while attempting to parse generated XML: {e}")

    assert parsed.find(".//categories/category[1]/name").text == "LU.3"



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import re
import typing
import pathlib
import unicodedata
from enum import Enum
from collections import UserString

from typing import (
    Any,
    Iterable,
    Sequence,
    TypeVar,
)


T = TypeVar("T")


def idem(obj: T, *args, **kwargs) -> T:
    """
    :returns: the parameter ``obj`` unchanged. Useful for some lambdas.
    """
    return obj


def get_openlane_root() -> str:
    """
    Returns the root OpenLane folder, i.e., the folder containing the
    ``__init__.py``.
    """
    return os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


def get_script_dir() -> str:
    """
    Gets the OpenLane tool `scripts` directory.

    :meta private:
    """
    return os.path.join(
        get_openlane_root(),
        "scripts",
    )


def get_opdks_rev() -> str:
    """
    Gets the Open_PDKs revision confirmed compatible with this version of OpenLane.
    """
    return open(
        os.path.join(get_openlane_root(), "open_pdks_rev"), encoding="utf8"
    ).read()


# The following code snippet has been adapted under the following license:
#
# Copyright (c) Django Software Foundation and individual contributors.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

#     1. Redistributions of source code must retain the above copyright notice,
#        this list of conditions and the following disclaimer.

#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.

#     3. Neither the name of Django nor the names of its contributors may be used
#        to endorse or promote products derived from this software without
#        specific prior written permission.


# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
def slugify(value: str) -> str:
    """
    :param value: Input string
    :returns: The input string converted to lower case, with all characters
        except alphanumerics, underscores and hyphens removed, and spaces and\
        dots converted into hyphens.

        Leading and trailing whitespace is stripped.
    """
    value = (
        unicodedata.normalize("NFKD", value).encode("ascii", "ignore").decode("ascii")
    )
    value = re.sub(r"[^\w\s\-\.]", "", value).strip().lower()
    return re.sub(r"[\s\.]+", "-", value)


def protected(method):
    """A decorator to indicate protected methods.

    It dynamically adds a statement to the effect in the docstring as well
    as setting an attribute, ``protected``, to ``True``, but has no other effects.

    :param f: Method to mark as protected
    """
    if method.__doc__ is None:
        method.__doc__ = ""
    method.__doc__ = "**protected**\n" + method.__doc__

    setattr(method, "protected", True)
    return method


final = typing.final


def mkdirp(path: typing.Union[str, os.PathLike]):
    """
    Attempts to create a directory and all of its parents.

    Does not fail if the directory already exists, however, it does fail
    if it is unable to create any of the components and/or if the path
    already exists as a file.

    :param path: A filesystem path for the directory
    """
    return pathlib.Path(path).mkdir(parents=True, exist_ok=True)


def StringEnum(name: str, values: Sequence[str]):
    """
    Creates a string enumeration class where the keys and values are the same.
    """
    return Enum(name, [(value, value) for value in values])


class Path(UserString, os.PathLike):
    """
    A Path type for OpenLane configuration variables.

    Basically just a string.
    """

    def __fspath__(self) -> str:
        return str(self)

    def exists(self) -> bool:
        """
        A convenience method calling :meth:`os.path.exists`
        """
        return os.path.exists(self)


class zip_first(object):
    """
    Works like ``zip_longest`` if |a| > |b| and ``zip`` if |a| <= |b|.
    """

    def __init__(self, a: Iterable, b: Iterable, fillvalue: Any) -> None:
        self.a = a
        self.b = b
        self.fillvalue = fillvalue

    def __iter__(self):
        self.iter_a = iter(self.a)
        self.iter_b = iter(self.b)
        return self

    def __next__(self):
        a = next(self.iter_a)
        b = self.fillvalue
        try:
            b = next(self.iter_b)
        except StopIteration:
            pass
        return (a, b)



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from shutil import rmtree
from unittest import mock
from decimal import Decimal
from typing import Any, Callable, Dict, Iterable, List, Optional

import pytest
from pyfakefs.fake_filesystem_unittest import Patcher

from . import config
from ..common import Path, StringEnum


@pytest.fixture()
def _mock_fs():
    with Patcher() as patcher:
        rmtree("/run", ignore_errors=True)
        patcher.fs.create_dir("/cwd/src")
        patcher.fs.create_file("/cwd/src/a.v")
        patcher.fs.create_file("/cwd/src/b.v")
        patcher.fs.create_dir("/cwd/spef")
        patcher.fs.create_file("/cwd/spef/b.spef")
        patcher.fs.create_file(
            "/pdk/dummy/libs.tech/openlane/config.tcl",
            contents="""
            if { ![info exists ::env(STD_CELL_LIBRARY)] } {
                set ::env(STD_CELL_LIBRARY) "dummy_scl"
            }
            set ::env(TECH_LEF) "/pdk/dummy/libs.ref/techlef/dummy_scl/dummy_tech_lef.tlef"
            set ::env(LIB_SYNTH) "sky130_fd_sc_hd__tt_025C_1v80.lib"
            """,
        )
        patcher.fs.create_file(
            "/pdk/dummy2/libs.tech/openlane/config.tcl",
            contents="""
            if { ![info exists ::env(STD_CELL_LIBRARY)] } {
                set ::env(STD_CELL_LIBRARY) "dummy2_scl"
            }
            set ::env(TECH_LEF) "/pdk/dummy2/libs.ref/techlef/dummy_scl/dummy_tech_lef.tlef"
            set ::env(LIB_SYNTH) "sky130_fd_sc_hd__tt_025C_1v80.lib"
            """,
        )
        patcher.fs.create_file(
            "/pdk/dummy/libs.ref/techlef/dummy_scl/dummy_tech_lef.tlef",
        )
        patcher.fs.create_file(
            "/pdk/dummy2/libs.ref/techlef/dummy_scl/dummy_tech_lef.tlef",
        )
        patcher.fs.create_file(
            "/pdk/dummy/libs.tech/openlane/dummy_scl/config.tcl",
            contents="",
        )
        patcher.fs.create_file(
            "/pdk/dummy2/libs.tech/openlane/dummy2_scl/config.tcl",
            contents="",
        )

        os.chdir("/cwd")
        yield


DiodeOnPortsEnum = StringEnum("Ports", ["none", "in"])
MOCK_PDK_VARS = [
    config.Variable(
        "PDK",
        str,
        description="x",
    ),
    config.Variable(
        "STD_CELL_LIBRARY",
        str,
        description="x",
    ),
    config.Variable(
        "EXAMPLE_PDK_VAR",
        Decimal,
        description="x",
        default=10.0,
    ),
    config.Variable(
        "TECH_LEFS",
        Dict[str, Path],
        description="x",
    ),
    config.Variable(
        "DEFAULT_CORNER",
        str,
        description="x",
        default="nom_tt_025C_1v80",
    ),
]
MOCK_FLOW_VARS = [
    config.Variable(
        "DESIGN_DIR",
        Path,
        "The directory of the design. Does not need to be provided explicitly.",
    ),
    config.Variable(
        "DESIGN_NAME",
        str,
        description="x",
    ),
    config.Variable(
        "VERILOG_FILES",
        List[Path],
        description="x",
    ),
    config.Variable(
        "GRT_REPAIR_ANTENNAS",
        bool,
        description="x",
        default=True,
    ),
    config.Variable(
        "RUN_HEURISTIC_DIODE_INSERTION",
        bool,
        description="x",
        default=False,
    ),
    config.Variable(
        "DIODE_ON_PORTS",
        DiodeOnPortsEnum,
        description="x",
        default="none",
    ),
    config.Variable(
        "MACROS",
        Optional[Dict[str, config.Macro]],
        description="x",
        default=None,
    ),
]


def mock_variables(patch_in_objects: Optional[Iterable[Any]] = None):
    from . import config

    if patch_in_objects is None:
        patch_in_objects = []
    patch_in_objects = patch_in_objects + [config]

    def decorator(f: Callable):
        for o in patch_in_objects:
            if hasattr(o, "pdk_variables"):
                f = mock.patch.object(
                    o,
                    "pdk_variables",
                    MOCK_PDK_VARS,
                )(f)
            if hasattr(o, "flow_common_variables"):
                f = mock.patch.object(
                    o,
                    "flow_common_variables",
                    MOCK_FLOW_VARS,
                )(f)
            if hasattr(o, "removed_variables"):
                f = mock.patch.object(
                    o,
                    "removed_variables",
                    {"REMOVED_VARIABLE": "Variable sucked"},
                )(f)
            if hasattr(o, "universal_flow_config_variables"):
                f = mock.patch.object(
                    o,
                    "universal_flow_config_variables",
                    MOCK_FLOW_VARS,
                )(f)
            if hasattr(o, "all_variables"):
                f = mock.patch.object(
                    o,
                    "all_variables",
                    MOCK_FLOW_VARS,
                )(f)

        return f

    return decorator


@pytest.mark.usefixtures("_mock_fs")
@mock_variables()
def test_dict_config():
    from . import Meta, Config

    cfg, _ = Config.load(
        {"DESIGN_NAME": "whatever", "VERILOG_FILES": "dir::src/*.v"},
        config.flow_common_variables,
        design_dir="/cwd",
        pdk="dummy",
        scl="dummy_scl",
        pdk_root="/pdk",
    )

    assert_cfg = Config(
        {
            "DESIGN_DIR": "/cwd",
            "DESIGN_NAME": "whatever",
            "PDK": "dummy",
            "STD_CELL_LIBRARY": "dummy_scl",
            "VERILOG_FILES": ["/cwd/src/a.v", "/cwd/src/b.v"],
            "EXAMPLE_PDK_VAR": Decimal("10"),
            "GRT_REPAIR_ANTENNAS": True,
            "RUN_HEURISTIC_DIODE_INSERTION": False,
            "DIODE_ON_PORTS": DiodeOnPortsEnum.none,
            "MACROS": None,
            "TECH_LEFS": {
                "nom_*": Path(
                    "/pdk/dummy/libs.ref/techlef/dummy_scl/dummy_tech_lef.tlef"
                )
            },
            "DEFAULT_CORNER": "nom_tt_025C_1v80",
        },
        meta=Meta(version=2, flow="Classic"),
    )
    assert cfg == assert_cfg, "Generated configuration does not match expected value"


@pytest.mark.usefixtures("_mock_fs")
@mock_variables()
def test_json_config():
    from . import Meta, Config

    with open("/cwd/config.json", "w") as f:
        f.write(
            """
            {
                "meta": {
                    "version": 2,
                    "flow": "Whatever"
                },
                "DESIGN_NAME": "whatever",
                "VERILOG_FILES": "dir::src/*.v"
            }
            """
        )

    cfg, _ = Config.load(
        "/cwd/config.json",
        config.flow_common_variables,
        pdk="dummy",
        scl="dummy_scl",
        pdk_root="/pdk",
    )

    assert cfg == Config(
        {
            "DESIGN_DIR": "/cwd",
            "DESIGN_NAME": "whatever",
            "PDK": "dummy",
            "STD_CELL_LIBRARY": "dummy_scl",
            "VERILOG_FILES": ["/cwd/src/a.v", "/cwd/src/b.v"],
            "EXAMPLE_PDK_VAR": Decimal("10"),
            "GRT_REPAIR_ANTENNAS": True,
            "RUN_HEURISTIC_DIODE_INSERTION": False,
            "DIODE_ON_PORTS": DiodeOnPortsEnum.none,
            "MACROS": None,
            "TECH_LEFS": {
                "nom_*": Path(
                    "/pdk/dummy/libs.ref/techlef/dummy_scl/dummy_tech_lef.tlef"
                )
            },
            "DEFAULT_CORNER": "nom_tt_025C_1v80",
        },
        meta=Meta(version=2, flow="Whatever"),
    ), "Generated configuration does not match expected value"

    assert Config.get_meta("/cwd/config.json", flow_override="OtherWhatever") == Meta(
        version=2, flow="OtherWhatever"
    ), "get_meta test failed"


@pytest.mark.usefixtures("_mock_fs")
@mock_variables()
def test_tcl_config():
    from . import Meta, Config

    with open("/cwd/config.tcl", "w") as f:
        f.write(
            """
            set ::env(DESIGN_NAME) "whatever"
            set ::env(VERILOG_FILES) "\\
                /cwd/src/a.v\\
                    /cwd/src/b.v\\
            "
            # cant test glob because of the mock filesystem
            """
        )

    cfg, _ = Config.load(
        "/cwd/config.tcl",
        config.flow_common_variables,
        pdk="dummy",
        scl="dummy_scl",
        pdk_root="/pdk",
    )

    assert cfg == Config(
        {
            "DESIGN_DIR": "/cwd",
            "DESIGN_NAME": "whatever",
            "PDK": "dummy",
            "STD_CELL_LIBRARY": "dummy_scl",
            "VERILOG_FILES": ["/cwd/src/a.v", "/cwd/src/b.v"],
            "EXAMPLE_PDK_VAR": Decimal("10"),
            "GRT_REPAIR_ANTENNAS": True,
            "RUN_HEURISTIC_DIODE_INSERTION": False,
            "DIODE_ON_PORTS": DiodeOnPortsEnum.none,
            "MACROS": None,
            "TECH_LEFS": {
                "nom_*": Path(
                    "/pdk/dummy/libs.ref/techlef/dummy_scl/dummy_tech_lef.tlef"
                )
            },
            "DEFAULT_CORNER": "nom_tt_025C_1v80",
        },
        meta=Meta(version=1, flow="Classic"),
    ), "Generated configuration does not match expected value"


@pytest.mark.usefixtures("_mock_fs")
@mock_variables()
def test_copy_filtered():
    from . import Config, Variable

    step1_variables = [
        Variable(
            "STEP1_VAR",
            int,
            description="x",
        )
    ]
    step2_variables = [
        Variable(
            "STEP2_VAR",
            int,
            description="x",
        )
    ]

    cfg, _ = Config.load(
        {
            "DESIGN_NAME": "whatever",
            "VERILOG_FILES": "dir::src/*.v",
            "STEP1_VAR": 3,
            "STEP2_VAR": 4,
        },
        config.flow_common_variables + step1_variables + step2_variables,
        design_dir="/cwd",
        pdk="dummy",
        scl="dummy_scl",
        pdk_root="/pdk",
    )

    step1_cfg = cfg.copy_filtered(step1_variables, include_pdk_variables=False)

    assert step1_cfg == {
        "DESIGN_DIR": "/cwd",
        "DESIGN_NAME": "whatever",
        "STEP1_VAR": 3,
        "VERILOG_FILES": ["/cwd/src/a.v", "/cwd/src/b.v"],
        "meta": cfg.meta,
        "GRT_REPAIR_ANTENNAS": True,
        "RUN_HEURISTIC_DIODE_INSERTION": False,
        "DIODE_ON_PORTS": DiodeOnPortsEnum.none,
        "MACROS": None,
    }, "copy_filtered for step 1 did not work properly"

    step2_cfg = cfg.copy_filtered(step2_variables, include_common_variables=False)

    assert step2_cfg == {
        "STEP2_VAR": 4,
        "PDK": "dummy",
        "STD_CELL_LIBRARY": "dummy_scl",
        "EXAMPLE_PDK_VAR": Decimal("10"),
        "meta": cfg.meta,
        "TECH_LEFS": {
            "nom_*": Path("/pdk/dummy/libs.ref/techlef/dummy_scl/dummy_tech_lef.tlef")
        },
        "DEFAULT_CORNER": "nom_tt_025C_1v80",
    }, "copy_filtered for step 2 did not work properly"


@pytest.mark.usefixtures("_mock_fs")
@mock_variables()
def test_automatic_conversion():
    from . import Config

    with open("/cwd/config.json", "w") as f:
        f.write(
            """
            {
                "DESIGN_NAME": "whatever",
                "VERILOG_FILES": "/cwd/src/a.v /cwd/src/b.v"
            }
            """
        )

    cfg, _ = Config.load(
        "/cwd/config.json",
        config.flow_common_variables,
        pdk="dummy",
        scl="dummy_scl",
        pdk_root="/pdk",
    )

    assert cfg["VERILOG_FILES"] == [
        "/cwd/src/a.v",
        "/cwd/src/b.v",
    ], "automatic conversion of tcl-style list failed for json file"

    with open("/cwd/config.json", "w") as f:
        f.write(
            """
            {
                "meta": {
                    "version": 2,
                    "flow": "Whatever"
                },
                "DESIGN_NAME": "whatever",
                "VERILOG_FILES": "/cwd/src/a.v /cwd/src/b.v"
            }
            """
        )

    with pytest.raises(
        config.InvalidConfig, match="Refusing to automatically convert"
    ) as e:
        Config.load(
            "/cwd/config.json",
            config.flow_common_variables,
            pdk="dummy",
            scl="dummy_scl",
            pdk_root="/pdk",
        )
    assert e is not None, "invalid config accepted: automatic conversion for meta >= 2"


@pytest.mark.usefixtures("_mock_fs")
@mock_variables()
def test_pdk():
    from . import InvalidConfig, Config

    cfg1, _ = Config.load(
        {
            "DESIGN_NAME": "whatever",
            "VERILOG_FILES": "dir::src/*.v",
        },
        config.flow_common_variables,
        pdk="dummy",
        design_dir="/cwd",
        pdk_root="/pdk",
    )
    assert cfg1["PDK"] == "dummy", "PDK in argument failed to load"

    cfg1, _ = Config.load(
        {
            "PDK": "dummy2",
            "DESIGN_NAME": "whatever",
            "VERILOG_FILES": "dir::src/*.v",
        },
        config.flow_common_variables,
        design_dir="/cwd",
        pdk_root="/pdk",
    )
    assert cfg1["PDK"] == "dummy2", "PDK in dictionary failed to load"

    cfg2, _ = Config.load(
        {
            "PDK": "dummy2",
            "DESIGN_NAME": "whatever",
            "VERILOG_FILES": "dir::src/*.v",
        },
        config.flow_common_variables,
        design_dir="/cwd",
        pdk_root="/pdk",
        pdk="dummy",
    )
    assert cfg2["PDK"] == "dummy2", "PDK in dictionary does not take priority"

    with pytest.raises(ValueError, match="pdk argument is required") as e:
        Config.load(
            {
                "DESIGN_NAME": "whatever",
                "VERILOG_FILES": "dir::src/*.v",
            },
            config.flow_common_variables,
            design_dir="/cwd",
            pdk_root="/pdk",
        )

    assert e is not None, "configuration loaded without PDK set"

    with pytest.raises(InvalidConfig, match="was not found") as e:
        Config.load(
            {
                "DESIGN_NAME": "whatever",
                "VERILOG_FILES": "dir::src/*.v",
            },
            config.flow_common_variables,
            design_dir="/cwd",
            pdk_root="/pdk",
            pdk="notreal",
        )

    assert e is not None, "configuration loaded without invalid PDK exception"

    with pytest.raises(InvalidConfig, match="not found") as e:
        Config.load(
            {
                "DESIGN_NAME": "whatever",
                "VERILOG_FILES": "dir::src/*.v",
            },
            config.flow_common_variables,
            design_dir="/cwd",
            pdk_root="/pdk",
            pdk="dumm",
        )

    assert True in [
        "similarly-named" in warning for warning in e.value.warnings
    ], "invalid PDK did not return suggestions in warnings"


@pytest.mark.usefixtures("_mock_fs")
@mock_variables()
def test_invalid_keys(caplog: pytest.LogCaptureFixture):
    from . import InvalidConfig, Config

    with open("/cwd/config.json", "w") as f:
        f.write(
            """
            {
                "DESIGN_NAME": "whatever",
                "VERILOG_FILES": "dir::src/*.v",
                "INVALID_VARIABLE": "potato"
            }
            """
        )

    try:
        Config.load(
            "/cwd/config.json",
            config.flow_common_variables,
            pdk="dummy",
            scl="dummy_scl",
            pdk_root="/pdk",
        )
    except InvalidConfig as e:
        assert (  # noqa: PT017
            e is None
        ), "unknown variable triggered an error when loading from a meta.version: 1 JSON file"

    assert (
        "Unknown key" in caplog.text
    ), "unknown variable did not trigger a warning when loading from a meta.version: 1 JSON file"
    caplog.clear()

    with pytest.raises(InvalidConfig, match="Unknown key") as e:
        Config.load(
            {
                "DESIGN_NAME": "whatever",
                "VERILOG_FILES": "dir::src/*.v",
                "INVALID_VARIABLE": "potato",
            },
            config.flow_common_variables,
            design_dir="/cwd",
            pdk="dummy",
            scl="dummy_scl",
            pdk_root="/pdk",
        )

    assert (
        e is not None
    ), "unknown variable did not trigger an error when loading from dict"

    Config.load(
        {
            "DESIGN_NAME": "whatever",
            "VERILOG_FILES": "dir::src/*.v",
            "REMOVED_VARIABLE": "potato",
        },
        config.flow_common_variables,
        design_dir="/cwd",
        pdk="dummy",
        scl="dummy_scl",
        pdk_root="/pdk",
    )

    assert (
        "has been removed" in caplog.text
    ), "removed variable did not trigger a warning when loading from a meta.version: 1 JSON file"
    caplog.clear()


@pytest.mark.usefixtures("_mock_fs")
@mock_variables()
def test_dis_migration(caplog: pytest.LogCaptureFixture):
    from . import Config, InvalidConfig

    def set_dis(dis: int):
        with open("/cwd/config.json", "w") as f:
            f.write(
                f"""
                {{
                    "DESIGN_NAME": "whatever",
                    "VERILOG_FILES": "dir::src/*.v",
                    "DIODE_INSERTION_STRATEGY": {dis}
                }}
                """
            )

    for dis in [1, 2, 5, 7, '"cat"']:
        set_dis(dis)
        with pytest.raises(InvalidConfig, match="not available in OpenLane") as e:
            Config.load(
                "/cwd/config.json",
                config.flow_common_variables,
                pdk="dummy",
                scl="dummy_scl",
                pdk_root="/pdk",
            )
        assert e is not None, f"configuration with invalid dis {dis} has been loaded"

    set_dis(0)
    cfg, _ = Config.load(
        "/cwd/config.json",
        config.flow_common_variables,
        pdk="dummy",
        scl="dummy_scl",
        pdk_root="/pdk",
    )

    assert not cfg["GRT_REPAIR_ANTENNAS"], "failed to migrate dis 0 properly"
    assert not cfg["RUN_HEURISTIC_DIODE_INSERTION"], "failed to migrate dis 0 properly"
    assert (
        cfg["DIODE_ON_PORTS"] == DiodeOnPortsEnum.none
    ), "failed to migrate dis 0 properly"

    set_dis(3)
    cfg, _ = Config.load(
        "/cwd/config.json",
        config.flow_common_variables,
        pdk="dummy",
        scl="dummy_scl",
        pdk_root="/pdk",
    )
    assert cfg["GRT_REPAIR_ANTENNAS"], "failed to migrate dis 3 properly"
    assert not cfg["RUN_HEURISTIC_DIODE_INSERTION"], "failed to migrate dis 3 properly"
    assert (
        cfg["DIODE_ON_PORTS"] == DiodeOnPortsEnum.none
    ), "failed to migrate dis 3 properly"

    set_dis(4)
    cfg, _ = Config.load(
        "/cwd/config.json",
        config.flow_common_variables,
        pdk="dummy",
        scl="dummy_scl",
        pdk_root="/pdk",
    )
    assert not cfg["GRT_REPAIR_ANTENNAS"], "failed to migrate dis 4 properly"
    assert cfg["RUN_HEURISTIC_DIODE_INSERTION"], "failed to migrate dis 4 properly"
    assert (
        cfg["DIODE_ON_PORTS"] == DiodeOnPortsEnum["in"]
    ), "failed to migrate dis 4 properly"

    set_dis(6)
    cfg, _ = Config.load(
        "/cwd/config.json",
        config.flow_common_variables,
        pdk="dummy",
        scl="dummy_scl",
        pdk_root="/pdk",
    )
    assert cfg["GRT_REPAIR_ANTENNAS"], "failed to migrate dis 6 properly"
    assert cfg["RUN_HEURISTIC_DIODE_INSERTION"], "failed to migrate dis 6 properly"
    assert (
        cfg["DIODE_ON_PORTS"] == DiodeOnPortsEnum["in"]
    ), "failed to migrate dis 6 properly"

    assert (
        "See 'Migrating DIODE_INSERTION_STRATEGY'" in caplog.text
    ), "diode insertion strategy did not trigger a warning"
    caplog.clear()


@pytest.mark.usefixtures("_mock_fs")
@mock_variables()
def test_macro_migration(
    caplog: pytest.LogCaptureFixture,
):
    from . import Config, Macro, InvalidConfig

    cfg, _ = Config.load(
        {
            "DESIGN_NAME": "whatever",
            "VERILOG_FILES": "dir::src/*.v",
            "EXTRA_SPEFS": "b /cwd/spef/b.spef /cwd/spef/b.spef /cwd/spef/b.spef",
        },
        config.flow_common_variables,
        design_dir="/cwd",
        pdk="dummy",
        scl="dummy_scl",
        pdk_root="/pdk",
    )

    assert cfg["MACROS"] == {
        "b": Macro(
            gds=[""],
            lef=[""],
            instances={},
            nl=[],
            spef={
                "min_*": ["/cwd/spef/b.spef"],
                "nom_*": ["/cwd/spef/b.spef"],
                "max_*": ["/cwd/spef/b.spef"],
            },
            lib={},
            spice=[],
            sdf={},
            json_h=None,
        )
    }, "Macro migration yielded unexpected result"

    assert (
        "deprecated" in caplog.text
    ), "configuration variable 'EXTRA_SPEFS' is deprecated"
    caplog.clear()

    with pytest.raises(InvalidConfig, match="not divisible by four") as e:
        Config.load(
            {
                "DESIGN_NAME": "whatever",
                "VERILOG_FILES": "dir::src/*.v",
                "EXTRA_SPEFS": "b /cwd/spef/b.spef /cwd/spef/b.spef",
            },
            config.flow_common_variables,
            design_dir="/cwd",
            pdk="dummy",
            scl="dummy_scl",
            pdk_root="/pdk",
        )

    assert e is not None, "invalid EXTRA_SPEFS count accepted"

    with pytest.raises(InvalidConfig, match="Invalid type for 'EXTRA_SPEFS'") as e:
        Config.load(
            {
                "DESIGN_NAME": "whatever",
                "VERILOG_FILES": "dir::src/*.v",
                "EXTRA_SPEFS": 2,
            },
            config.flow_common_variables,
            design_dir="/cwd",
            pdk="dummy",
            scl="dummy_scl",
            pdk_root="/pdk",
        )

    assert e is not None, "invalid EXTRA_SPEFS type accepted"



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest
from pyfakefs.fake_filesystem import FakeFilesystem
from ..config.test_config import (  # noqa: F401
    _mock_fs,
)


def test_cli_basic():
    import click
    from .cli import cloup_flow_opts

    @click.command()
    @cloup_flow_opts(volare_by_default=False)
    def cli_fn(**kwargs):
        return kwargs

    with pytest.raises(SystemExit):
        cli_fn([])


def test_only_flag():
    import click
    from .cli import cloup_flow_opts

    @click.command()
    @cloup_flow_opts(volare_by_default=False)
    def cli_fn(**kwargs):
        return kwargs

    result = cli_fn(
        ["--only", "this-step"],
        standalone_mode=False,
    )
    assert result["frm"] == "this-step", "only flag not translated to frm correctly"
    assert result["to"] == "this-step", "only flag not translated to to correctly"
    assert (
        "only" not in result
    ), "only flag not disposed of after translating to from and to"


def test_log_level_flag(caplog: pytest.LogCaptureFixture):
    import click
    from .cli import cloup_flow_opts

    from ..logging import get_log_level, reset_log_level

    @click.command()
    @cloup_flow_opts(volare_by_default=False)
    def cli_fn(**kwargs):
        return kwargs

    cli_fn(
        ["--log-level", "30"],
        standalone_mode=False,
    )
    assert get_log_level() == 30, "--log-level callback failed"

    reset_log_level()

    cli_fn(
        ["--log-level", "WARN"],
        standalone_mode=False,
    )
    assert get_log_level() == 30, "--log-level callback failed"

    reset_log_level()

    cli_fn(
        ["--log-level", "NOT A REAL LOG LEVEL"],
        standalone_mode=False,
    )
    assert (
        "Invalid logging level" in caplog.text
    ), "Invalid logging level did not raise an error"
    caplog.clear()


def test_worker_count_cb():
    import click
    from .cli import cloup_flow_opts

    from ..common import get_tpe, set_tpe

    @click.command()
    @cloup_flow_opts(volare_by_default=False)
    def cli_fn(**kwargs):
        return kwargs

    tpe_backup = get_tpe()

    cli_fn(
        ["-j", "3"],
        standalone_mode=False,
    )
    assert get_tpe()._max_workers == 3, "--jobs callback failed"

    set_tpe(tpe_backup)


pytest.mark.usefixtures("_mock_fs")


def test_initial_state(fs: FakeFilesystem, caplog: pytest.LogCaptureFixture):
    import click
    from .cli import cloup_flow_opts

    @click.command()
    @cloup_flow_opts(volare_by_default=False)
    def cli_fn(**kwargs):
        return kwargs

    fs.create_file(
        "/cwd/state_in.json",
        contents="""
        {
            "metrics": {"hi": true}
        }
        """,
    )

    fs.create_file(
        "/cwd/bad_json.json",
        contents="""
        {
            "metrics": {"hi": true}
        """,
    )

    fs.create_file(
        "/cwd/non_dict.json",
        contents="""
        []
        """,
    )

    results = cli_fn(
        ["--with-initial-state", "/cwd/state_in.json"],
        standalone_mode=False,
    )

    assert results["with_initial_state"].metrics["hi"], "Failed to load initial state"

    cli_fn(
        ["--with-initial-state", "/cwd/bad_json.json"],
        standalone_mode=False,
    )
    assert "Invalid JSON" in caplog.text, "Invalid JSON file did not report an error"
    caplog.clear()

    cli_fn(
        ["--with-initial-state", "/cwd/non_dict.json"],
        standalone_mode=False,
    )
    assert (
        "is not a dictionary" in caplog.text
    ), "Non-dictionary JSON file did not report an error"
    caplog.clear()



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import math
import enum
from decimal import Decimal
from collections import UserString
from concurrent.futures import ThreadPoolExecutor

import pytest


class MyString(UserString):
    pass


def test_slugify():
    from . import slugify

    assert slugify("ABCD efg.xy-Z") == "abcd-efg-xy-z", "Failed slugify test"
    assert (
        slugify("Lorem ipsum   dolor sit amet") == "lorem-ipsum-dolor-sit-amet"
    ), "Failed slugify test"


def test_stringenum():
    from . import StringEnum

    breeds = ["Siamese", "Persian", "Ragdoll", "Maine Coon"]

    with pytest.raises(TypeError, match="Attempted to reuse key"):
        StringEnum("CatBreeds", breeds + ["Siamese"])

    valid_enum = StringEnum("CatBreeds", breeds)

    for breed in breeds:
        assert (
            getattr(valid_enum, breed).value == breed
        ), f"Enum member '{breed}' did not get a matching string"
        assert (
            valid_enum[breed].value == breed
        ), f"Enum member '{breed}' could not be properly indexed"


def test_is_string():
    from . import is_string

    assert is_string(
        "just a normal string"
    ), "is_string is not accepting a python string"

    assert not is_string(b"a byte string"), "is_string is accepting a byte string"

    assert is_string(
        MyString("a userstring")
    ), "is_string is not accepting a userstring"


def test_parse_metric_modifiers():
    from . import parse_metric_modifiers

    assert parse_metric_modifiers("category__name__optional_name_modifier__etc",) == (
        "category__name__optional_name_modifier__etc",
        {},
    ), "Improperly parsed metric without modifiers"

    assert parse_metric_modifiers(
        "category__name__optional_name_modifier__etc__mod1:one:mod2:two"
    ) == (
        "category__name__optional_name_modifier__etc",
        {"mod1": "one", "mod2": "two"},
    ), "Improperly parsed metric with modifiers"

    assert parse_metric_modifiers(
        "category__name__optional_name_modifier__etc:etc:etc",
    ) == (
        "category__name__optional_name_modifier__etc:etc:etc",
        {},
    ), "Improperly parsed metric with improper modifier syntax"


@pytest.mark.parametrize(
    ("input", "aggregators", "expected"),
    [
        (
            {
                "flower__count__type:roses": 12,
                "flower__max__height__type:roses": Decimal("7.0"),
                "flower__count__type:tulips": 21,
                "flower__max__height__type:tulips": Decimal("8.0"),
                "flower__count": 0,
            },
            {
                "flower__count": (0, lambda x: sum(x)),
                "flower__max__height": (-math.inf, lambda x: max(x)),
            },
            {
                "flower__count__type:roses": 12,
                "flower__count__type:tulips": 21,
                "flower__count": 33,
                "flower__max__height__type:roses": Decimal("7.0"),
                "flower__max__height__type:tulips": Decimal("8.0"),
                "flower__max__height": Decimal("8.0"),
            },
        ),
    ],
)
def test_aggregate_metrics(input, aggregators, expected):
    from . import aggregate_metrics

    assert (
        aggregate_metrics(input, aggregators) == expected
    ), "aggregate_metrics() returned unexpected output"


def test_generic_dict():
    from . import GenericDict

    test_dict = GenericDict({"a": "b", "c": "d"}, overrides={"c": "e"})
    assert test_dict["a"] == "b", "Copying in constructor not working properly"
    assert test_dict["c"] == "e", "Overrides not working properly"

    assert test_dict.check("c") == (
        "c",
        "e",
    ), ".check not finding existing key/value pair"
    assert test_dict.check("f") == (
        None,
        None,
    ), ".check not finding existing key/value pair"

    test_dict.update({"a": "g"})
    assert test_dict["a"] == "g", ".update not updating values"


def test_immutable_generic_dict():
    from . import GenericImmutableDict

    test_dict = GenericImmutableDict({"a": "b", "c": "d"}, overrides={"c": "e"})
    assert test_dict["a"] == "b", "Copying in constructor not working properly"
    assert test_dict["c"] == "e", "Overrides not working properly"

    with pytest.raises(TypeError, match="is immutable"):
        test_dict.update({"a": "g"})


class MyEnum(enum.Enum):
    A = 4
    B = "Horse"
    Horse = "B"


deep_dict = {
    "a": [
        {
            "b": MyEnum.Horse,
            "d": Decimal(0.2),
            "f": ["g", "h"],
        },
    ],
    "i": {"j": ["k", MyString("l")]},
}


def test_generic_dict_encoder():
    from . import GenericDict

    assert (
        GenericDict(deep_dict).dumps(indent=0).replace("\n", "")
        == '{"a": [{"b": "Horse","d": 0.2,"f": ["g","h"]}],"i": {"j": ["k","l"]}}'
    ), "Failed to serialize deep dictionary"

    assert (
        GenericDict(deep_dict).dumps(indent=1).replace("\n", "")
        == '{ "a": [  {   "b": "Horse",   "d": 0.2,   "f": [    "g",    "h"   ]  } ], "i": {  "j": [   "k",   "l"  ] }}'
    ), "Failed to properly handle indent kwarg"


def test_copy_recursive():
    from . import copy_recursive

    deep_dict_copy = copy_recursive(deep_dict)

    assert id(deep_dict["a"][0]["f"]) != id(
        deep_dict_copy["a"][0]["f"]
    ), "Reference in list identical to original"

    assert id(deep_dict["i"]["j"]) != id(
        deep_dict_copy["i"]["j"]
    ), "Reference in dict identical to original"

    deep_dict_copy["x"] = deep_dict_copy

    with pytest.raises(ValueError, match="Circular"):
        copy_recursive(deep_dict_copy)


def test_tpe():
    from . import get_tpe, set_tpe

    tpe = get_tpe()
    assert tpe._max_workers == os.cpu_count(), "TPE was not initialized properly"

    tpe = ThreadPoolExecutor(1)
    set_tpe(tpe)

    assert get_tpe() == tpe, "Failed to set TPE properly"


def test_immutable_dict():
    from . import GenericImmutableDict

    immutable_dict = GenericImmutableDict({"a": "d", "p": 4})

    with pytest.raises(TypeError, match="is immutable") as e:
        immutable_dict["p"] = 9

    assert e is not None, "Was able to mutate immutable dict"

    with pytest.raises(TypeError, match="is immutable") as e:
        del immutable_dict["a"]

    assert e is not None, "Was able to delete from immutable dict"



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
RAW_SKY130A_IN = {
    "PDK_ROOT": "/pdk_root",
    "PDK": "sky130A",
    "CTS_TECH_DIR": "N/A",
    "FP_PDN_HOFFSET": "16.65",
    "GPIO_PADS_VERILOG": " /pdk_root/sky130A/libs.ref/sky130_fd_io/verilog/sky130_ef_io.v\n",
    "STD_CELL_POWER_PINS": "VPWR VPB",
    "CELLS_LEF_OPT": "/pdk_root/sky130A/libs.ref/sky130_fd_sc_hd/lef/sky130_ef_sc_hd.lef /pdk_root/sky130A/libs.ref/sky130_fd_sc_hd/lef/sky130_fd_sc_hd.lef",
    "FP_IO_VLAYER": "met2",
    "MAGIC_TECH_FILE": "/pdk_root/sky130A/libs.tech/magic/sky130A.tech",
    "FP_PDN_VERTICAL_LAYER": "met4",
    "GRT_LAYER_ADJUSTMENTS": "0.99,0,0,0,0,0",
    "FP_PDN_CORE_RING_VWIDTH": "1.6",
    "FP_PDN_HSPACING": "1.7",
    "WIRE_RC_LAYER": "met1",
    "FP_PDN_VOFFSET": "16.32",
    "TECH_LEF_MAX": "/pdk_root/sky130A/libs.ref/sky130_fd_sc_hd/techlef/sky130_fd_sc_hd__max.tlef",
    "KLAYOUT_DEF_LAYER_MAP": "/pdk_root/sky130A/libs.tech/klayout/tech/sky130A.map",
    "MAGIC_MAGICRC": "/pdk_root/sky130A/libs.tech/magic/sky130A.magicrc",
    "FULL_ADDER_MAP": "/pdk_root/sky130A/libs.tech/openlane/sky130_fd_sc_hd/fa_map.v",
    "CTS_SQR_RES": "0.125",
    "RT_MAX_LAYER": "met5",
    "STD_CELL_LIBRARY_OPT_CDL": "/pdk_root/sky130A/libs.ref/sky130_fd_sc_hd/cdl/sky130_fd_sc_hd.cdl",
    "VDD_PIN": "VPWR",
    "KLAYOUT_XOR_IGNORE_LAYERS": "81/14",
    "RCX_RULES_MAX": "/pdk_root/sky130A/libs.tech/openlane/rules.openrcx.sky130A.max.calibre",
    "TECH_LEF_OPT": "/pdk_root/sky130A/libs.ref/sky130_fd_sc_hd/techlef/sky130_fd_sc_hd__nom.tlef",
    "LIB_SLOWEST_OPT": "/pdk_root/sky130A/libs.ref/sky130_fd_sc_hd/lib/sky130_fd_sc_hd__ss_100C_1v60.lib",
    "CTS_SQR_CAP": "0.258e-3",
    "GDS_FILES_OPT": "/pdk_root/sky130A/libs.ref/sky130_fd_sc_hd/gds/sky130_fd_sc_hd.gds",
    "FP_PDN_HWIDTH": "1.6",
    "STD_CELL_GROUND_PINS": "VGND VNB",
    "TRACKS_INFO_FILE": "/pdk_root/sky130A/libs.tech/openlane/sky130_fd_sc_hd/tracks.info",
    "TECH_LEF": "/pdk_root/sky130A/libs.ref/sky130_fd_sc_hd/techlef/sky130_fd_sc_hd__nom.tlef",
    "TECH_LEF_MIN": "/pdk_root/sky130A/libs.ref/sky130_fd_sc_hd/techlef/sky130_fd_sc_hd__min.tlef",
    "FP_PDN_HPITCH": "153.18",
    "RCX_RULES": "/pdk_root/sky130A/libs.tech/openlane/rules.openrcx.sky130A.nom.calibre",
    "GPIO_PADS_PREFIX": "sky130_fd_io sky130_ef_io",
    "STD_CELL_LIBRARY": "sky130_fd_sc_hd",
    "FP_PDN_CORE_RING_VSPACING": "1.7",
    "KLAYOUT_PROPERTIES": "/pdk_root/sky130A/libs.tech/klayout/tech/sky130A.lyp",
    "DRC_EXCLUDE_CELL_LIST": "/pdk_root/sky130A/libs.tech/openlane/sky130_fd_sc_hd/drc_exclude.cells",
    "CARRY_SELECT_ADDER_MAP": "/pdk_root/sky130A/libs.tech/openlane/sky130_fd_sc_hd/csa_map.v",
    "NO_SYNTH_CELL_LIST": "/pdk_root/sky130A/libs.tech/openlane/sky130_fd_sc_hd/no_synth.cells",
    "GPIO_PADS_LEF": " /pdk_root/sky130A/libs.ref/sky130_fd_io/lef/sky130_fd_io.lef /pdk_root/sky130A/libs.ref/sky130_fd_io/lef/sky130_ef_io.lef ",
    "RIPPLE_CARRY_ADDER_MAP": "/pdk_root/sky130A/libs.tech/openlane/sky130_fd_sc_hd/rca_map.v",
    "CVC_SCRIPTS_DIR": "/pdk_root/sky130A/libs.tech/openlane/cvc",
    "CLOCK_WIRE_RC_LAYER": "met5",
    "RCX_RULES_MIN": "/pdk_root/sky130A/libs.tech/openlane/rules.openrcx.sky130A.min.calibre",
    "FP_TAPCELL_DIST": "13",
    "SYNTH_LATCH_MAP": "/pdk_root/sky130A/libs.tech/openlane/sky130_fd_sc_hd/latch_map.v",
    "FP_PDN_VWIDTH": "1.6",
    "MACRO_BLOCKAGES_LAYER": "li1 met1 met2 met3 met4",
    "STD_CELL_LIBRARY_CDL": "/pdk_root/sky130A/libs.ref/sky130_fd_sc_hd/cdl/sky130_fd_sc_hd.cdl",
    "RT_MIN_LAYER": "met1",
    "VDD_PIN_VOLTAGE": "1.80",
    "FP_PDN_VPITCH": "153.6",
    "DRC_EXCLUDE_CELL_LIST_OPT": "/pdk_root/sky130A/libs.tech/openlane/sky130_fd_sc_hd/drc_exclude.cells",
    "GDS_FILES": "/pdk_root/sky130A/libs.ref/sky130_fd_sc_hd/gds/sky130_fd_sc_hd.gds",
    "FP_PDN_RAIL_LAYER": "met1",
    "FP_PDN_HORIZONTAL_LAYER": "met5",
    "KLAYOUT_TECH": "/pdk_root/sky130A/libs.tech/klayout/tech/sky130A.lyt",
    "FP_PDN_CORE_RING_HOFFSET": "6",
    "NETGEN_SETUP_FILE": "/pdk_root/sky130A/libs.tech/netgen/sky130A_setup.tcl",
    "KLAYOUT_DRC_TECH_SCRIPT": "/pdk_root/sky130A/libs.tech/klayout/drc/sky130A_mr.drc",
    "FP_PDN_VSPACING": "1.7",
    "GND_PIN": "VGND",
    "CELLS_LEF": "/pdk_root/sky130A/libs.ref/sky130_fd_sc_hd/lef/sky130_ef_sc_hd.lef /pdk_root/sky130A/libs.ref/sky130_fd_sc_hd/lef/sky130_fd_sc_hd.lef",
    "TRISTATE_BUFFER_MAP": "/pdk_root/sky130A/libs.tech/openlane/sky130_fd_sc_hd/tribuff_map.v",
    "FP_PDN_RAIL_OFFSET": "0",
    "GPIO_PADS_LEF_CORE_SIDE": " /pdk_root/sky130A/libs.tech/openlane/custom_cells/lef/sky130_fd_io_core.lef /pdk_root/sky130A/libs.tech/openlane/custom_cells/lef/sky130_ef_io_core.lef ",
    "FP_IO_HLAYER": "met3",
    "FP_PDN_CORE_RING_HSPACING": "1.7",
    "STD_CELL_LIBRARY_OPT": "sky130_fd_sc_hd",
    "GND_PIN_VOLTAGE": "0.00",
    "FP_PDN_CORE_RING_HWIDTH": "1.6",
    "DEF_UNITS_PER_MICRON": "1000",
    "DATA_WIRE_RC_LAYER": "met2",
    "FP_PDN_CORE_RING_VOFFSET": "6",
    "PROCESS": "130",
    "PL_LIB": "/pdk_root/sky130A/libs.ref/sky130_fd_sc_hd/lib/sky130_fd_sc_hd__tt_025C_1v80.lib",
    "CELL_PAD_EXCLUDE": "sky130_fd_sc_hd__tap* sky130_fd_sc_hd__decap* sky130_ef_sc_hd__decap* sky130_fd_sc_hd__fill*",
    "SYNTH_MIN_BUF_PORT": "sky130_fd_sc_hd__buf_2 A X",
    "DIODE_CELL": "sky130_fd_sc_hd__diode_2",
    "LIB_SYNTH": "/pdk_root/sky130A/libs.ref/sky130_fd_sc_hd/lib/sky130_fd_sc_hd__tt_025C_1v80.lib",
    "CLK_BUFFER_INPUT": "A",
    "SYNTH_TIELO_PORT": "sky130_fd_sc_hd__conb_1 LO",
    "FP_PDN_RAIL_WIDTH": "0.48",
    "FP_WELLTAP_CELL": "sky130_fd_sc_hd__tapvpwrvgnd_1",
    "SYNTH_MUX4_MAP": "/pdk_root/sky130A/libs.tech/openlane/sky130_fd_sc_hd/mux4_map.v",
    "LIB_TYPICAL": "/pdk_root/sky130A/libs.ref/sky130_fd_sc_hd/lib/sky130_fd_sc_hd__tt_025C_1v80.lib",
    "FILL_CELL": "sky130_fd_sc_hd__fill*",
    "LIB_SLOWEST": "/pdk_root/sky130A/libs.ref/sky130_fd_sc_hd/lib/sky130_fd_sc_hd__ss_100C_1v60.lib",
    "DPL_CELL_PADDING": "0",
    "DIODE_CELL_PIN": "DIODE",
    "SYNTH_MUX_MAP": "/pdk_root/sky130A/libs.tech/openlane/sky130_fd_sc_hd/mux2_map.v",
    "CTS_CLK_BUFFER_LIST": "sky130_fd_sc_hd__clkbuf_8 sky130_fd_sc_hd__clkbuf_4 sky130_fd_sc_hd__clkbuf_2",
    "FAKEDIODE_CELL": "sky130_ef_sc_hd__fakediode_2",
    "MAX_FANOUT_CONSTRAINT": "10",
    "RE_BUFFER_CELL": "sky130_fd_sc_hd__buf_4",
    "SYNTH_TIEHI_PORT": "sky130_fd_sc_hd__conb_1 HI",
    "GPL_CELL_PADDING": "0",
    "CLK_BUFFER": "sky130_fd_sc_hd__clkbuf_4",
    "CLK_BUFFER_OUTPUT": "X",
    "ROOT_CLK_BUFFER": "sky130_fd_sc_hd__clkbuf_16",
    "SYNTH_DRIVING_CELL": "sky130_fd_sc_hd__inv_2",
    "SYNTH_DRIVING_CELL_PIN": "Y",
    "OUTPUT_CAP_LOAD": "33.442",
    "TRISTATE_CELL_PREFIX": "sky130_fd_sc_hd__ebuf",
    "LIB_FASTEST": "/pdk_root/sky130A/libs.ref/sky130_fd_sc_hd/lib/sky130_fd_sc_hd__ff_n40C_1v95.lib",
    "PLACE_SITE_WIDTH": "0.460",
    "PLACE_SITE": "unithd",
    "MAX_TRANSITION_CONSTRAINT": "0.75",
    "FP_ENDCAP_CELL": "sky130_fd_sc_hd__decap_3",
    "PLACE_SITE_HEIGHT": "2.720",
    "DECAP_CELL": "sky130_ef_sc_hd__decap_12 sky130_fd_sc_hd__decap_8 sky130_fd_sc_hd__decap_6 sky130_fd_sc_hd__decap_4 sky130_fd_sc_hd__decap_3",
    "CTS_ROOT_BUFFER": "sky130_fd_sc_hd__clkbuf_16",
    "CTS_MAX_CAP": "1.53169",
    "CELL_CLK_PORT": "CLK",
}

RAW_GF180MCUC_IN = {
    "PDK_ROOT": "/pdk_root",
    "PDK": "gf180mcuC",
    "FP_PDN_HOFFSET": "16.65",
    "GPIO_PADS_VERILOG": " /pdk_root/gf180mcuC/libs.ref/gf180mcu_io/verilog/GF018green_ipio_5p0c_75_5lm.v\n",
    "STD_CELL_POWER_PINS": "VDD VNW",
    "FP_IO_VLAYER": "Metal2",
    "MAGIC_TECH_FILE": "/pdk_root/gf180mcuC/libs.tech/magic/gf180mcuC.tech",
    "FP_PDN_VERTICAL_LAYER": "Metal4",
    "GRT_LAYER_ADJUSTMENTS": "0,0,0,0,0",
    "FP_PDN_CORE_RING_VWIDTH": "1.6",
    "FP_PDN_HSPACING": "1.7",
    "WIRE_RC_LAYER": "Metal2",
    "FP_PDN_VOFFSET": "16.32",
    "LIB_SYNTH": "/pdk_root/gf180mcuC/libs.ref/gf180mcu_fd_sc_mcu7t5v0/liberty/gf180mcu_fd_sc_mcu7t5v0__tt_025C_5v00.lib",
    "TECH_LEF_MAX": "/pdk_root/gf180mcuC/libs.ref/gf180mcu_fd_sc_mcu7t5v0/techlef/gf180mcu_fd_sc_mcu7t5v0__max.tlef",
    "KLAYOUT_DEF_LAYER_MAP": "/pdk_root/gf180mcuC/libs.tech/klayout/tech/gf180mcu.map",
    "MAGIC_MAGICRC": "/pdk_root/gf180mcuC/libs.tech/magic/gf180mcuC.magicrc",
    "FULL_ADDER_MAP": "/pdk_root/gf180mcuC/libs.tech/openlane/gf180mcu_fd_sc_mcu7t5v0/fa_map.v",
    "RT_MAX_LAYER": "Metal5",
    "VDD_PIN": "VDD",
    "LIB_TYPICAL": "/pdk_root/gf180mcuC/libs.ref/gf180mcu_fd_sc_mcu7t5v0/liberty/gf180mcu_fd_sc_mcu7t5v0__tt_025C_5v00.lib",
    "LIB_SLOWEST": "/pdk_root/gf180mcuC/libs.ref/gf180mcu_fd_sc_mcu7t5v0/liberty/gf180mcu_fd_sc_mcu7t5v0__ss_125C_4v50.lib",
    "RCX_RULES_MAX": "/pdk_root/gf180mcuC/libs.tech/openlane/rules.openrcx.gf180mcuC.max",
    "FP_PDN_HWIDTH": "1.6",
    "DRT_MIN_LAYER": "Metal1",
    "STD_CELL_GROUND_PINS": "VSS VPW",
    "TRACKS_INFO_FILE": "/pdk_root/gf180mcuC/libs.tech/openlane/gf180mcu_fd_sc_mcu7t5v0/tracks.info",
    "TECH_LEF": "/pdk_root/gf180mcuC/libs.ref/gf180mcu_fd_sc_mcu7t5v0/techlef/gf180mcu_fd_sc_mcu7t5v0__nom.tlef",
    "TECH_LEF_MIN": "/pdk_root/gf180mcuC/libs.ref/gf180mcu_fd_sc_mcu7t5v0/techlef/gf180mcu_fd_sc_mcu7t5v0__min.tlef",
    "FP_PDN_HPITCH": "153.18",
    "RCX_RULES": "/pdk_root/gf180mcuC/libs.tech/openlane/rules.openrcx.gf180mcuC.nom",
    "STD_CELL_LIBRARY": "gf180mcu_fd_sc_mcu7t5v0",
    "FP_PDN_CORE_RING_VSPACING": "1.7",
    "KLAYOUT_PROPERTIES": "/pdk_root/gf180mcuC/libs.tech/klayout/tech/gf180mcu.lyp",
    "DRC_EXCLUDE_CELL_LIST": "/pdk_root/gf180mcuC/libs.tech/openlane/gf180mcu_fd_sc_mcu7t5v0/drc_exclude.cells",
    "CARRY_SELECT_ADDER_MAP": "/pdk_root/gf180mcuC/libs.tech/openlane/gf180mcu_fd_sc_mcu7t5v0/csa_map.v",
    "NO_SYNTH_CELL_LIST": "/pdk_root/gf180mcuC/libs.tech/openlane/gf180mcu_fd_sc_mcu7t5v0/no_synth.cells",
    "GPIO_PADS_LEF": " /pdk_root/gf180mcuC/libs.ref/gf180mcu_io/lef/GF018green_ipio_5p0c_75_5lm.lef\n",
    "RIPPLE_CARRY_ADDER_MAP": "/pdk_root/gf180mcuC/libs.tech/openlane/gf180mcu_fd_sc_mcu7t5v0/rca_map.v",
    "RCX_RULES_MIN": "/pdk_root/gf180mcuC/libs.tech/openlane/rules.openrcx.gf180mcuC.min",
    "CLOCK_WIRE_RC_LAYER": "Metal4",
    "FP_TAPCELL_DIST": "20",
    "SYNTH_LATCH_MAP": "/pdk_root/gf180mcuC/libs.tech/openlane/gf180mcu_fd_sc_mcu7t5v0/latch_map.v",
    "FP_PDN_VWIDTH": "1.6",
    "MACRO_BLOCKAGES_LAYER": "Metal1 Metal2 Metal3 Metal4 Metal5",
    "STD_CELL_LIBRARY_CDL": "/pdk_root/gf180mcuC/libs.ref/gf180mcu_fd_sc_mcu7t5v0/cdl/gf180mcu_fd_sc_mcu7t5v0.cdl",
    "RT_MIN_LAYER": "Metal2",
    "VDD_PIN_VOLTAGE": "5.00",
    "FP_PDN_VPITCH": "153.6",
    "DRC_EXCLUDE_CELL_LIST_OPT": "/pdk_root/gf180mcuC/libs.tech/openlane/gf180mcu_fd_sc_mcu7t5v0/drc_exclude.cells",
    "GDS_FILES": "/pdk_root/gf180mcuC/libs.ref/gf180mcu_fd_sc_mcu7t5v0/gds/gf180mcu_fd_sc_mcu7t5v0.gds",
    "FP_PDN_RAIL_LAYER": "Metal1",
    "KLAYOUT_TECH": "/pdk_root/gf180mcuC/libs.tech/klayout/tech/gf180mcu.lyt",
    "FP_PDN_HORIZONTAL_LAYER": "Metal5",
    "FP_PDN_CORE_RING_HOFFSET": "6",
    "NETGEN_SETUP_FILE": "/pdk_root/gf180mcuC/libs.tech/netgen/gf180mcuC_setup.tcl",
    "KLAYOUT_DRC_TECH_SCRIPT": "/pdk_root/gf180mcuC/libs.tech/klayout/drc/gf180mcu.drc",
    "FP_PDN_VSPACING": "1.7",
    "GND_PIN": "VSS",
    "LIB_FASTEST": "/pdk_root/gf180mcuC/libs.ref/gf180mcu_fd_sc_mcu7t5v0/liberty/gf180mcu_fd_sc_mcu7t5v0__ff_n40C_5v50.lib",
    "CELLS_LEF": "/pdk_root/gf180mcuC/libs.ref/gf180mcu_fd_sc_mcu7t5v0/lef/gf180mcu_fd_sc_mcu7t5v0.lef",
    "TRISTATE_BUFFER_MAP": "/pdk_root/gf180mcuC/libs.tech/openlane/gf180mcu_fd_sc_mcu7t5v0/tribuff_map.v",
    "FP_PDN_RAIL_OFFSET": "0",
    "FP_IO_HLAYER": "Metal3",
    "FP_PDN_CORE_RING_HSPACING": "1.7",
    "GND_PIN_VOLTAGE": "0.00",
    "STD_CELL_LIBRARY_OPT": "gf180mcu_fd_sc_mcu7t5v0",
    "FP_PDN_CORE_RING_HWIDTH": "1.6",
    "DEF_UNITS_PER_MICRON": "2000",
    "DATA_WIRE_RC_LAYER": "Metal2",
    "METAL_LAYER_NAMES": "Metal1 Metal2 Metal3 Metal4 Metal5",
    "FP_PDN_CORE_RING_VOFFSET": "6",
    "PROCESS": "180",
    "PL_LIB": "/pdk_root/gf180mcuC/libs.ref/gf180mcu_fd_sc_mcu7t5v0/liberty/gf180mcu_fd_sc_mcu7t5v0__tt_025C_5v00.lib",
    "CELL_PAD_EXCLUDE": "gf180mcu_fd_sc_mcu7t5v0__filltie_* gf180mcu_fd_sc_mcu7t5v0__filldecap_* gf180mcu_fd_sc_mcu7t5v0__fill_* gf180mcu_fd_sc_mcu7t5v0__endcap_*",
    "SYNTH_MIN_BUF_PORT": "gf180mcu_fd_sc_mcu7t5v0__buf_1 I Z",
    "DIODE_CELL": "gf180mcu_fd_sc_mcu7t5v0__antenna",
    "SYNTH_CLK_DRIVING_CELL_PIN": "ZN",
    "SYNTH_TIELO_PORT": "gf180mcu_fd_sc_mcu7t5v0__tiel ZN",
    "FP_PDN_RAIL_WIDTH": "0.6",
    "FP_WELLTAP_CELL": "gf180mcu_fd_sc_mcu7t5v0__filltie",
    "FILL_CELL": "gf180mcu_fd_sc_mcu7t5v0__fill_*",
    "DPL_CELL_PADDING": "0",
    "DIODE_CELL_PIN": "I",
    "CTS_CLK_BUFFER_LIST": "gf180mcu_fd_sc_mcu7t5v0__clkbuf_2 gf180mcu_fd_sc_mcu7t5v0__clkbuf_4 gf180mcu_fd_sc_mcu7t5v0__clkbuf_8",
    "MAX_FANOUT_CONSTRAINT": "10",
    "SYNTH_CLK_DRIVING_CELL": "gf180mcu_fd_sc_mcu7t5v0__inv_4",
    "SYNTH_TIEHI_PORT": "gf180mcu_fd_sc_mcu7t5v0__tieh Z",
    "GPL_CELL_PADDING": "0",
    "SYNTH_DRIVING_CELL": "gf180mcu_fd_sc_mcu7t5v0__inv_1",
    "SYNTH_DRIVING_CELL_PIN": "ZN",
    "OUTPUT_CAP_LOAD": "72.91",
    "TRISTATE_CELL_PREFIX": "gf180mcu_fd_sc_mcu7t5v0__bufz",
    "DIODE_INSERTION_STRATEGY": "4",
    "PLACE_SITE_WIDTH": "0.56",
    "PLACE_SITE": "GF018hv5v_mcu_sc7",
    "MAX_TRANSITION_CONSTRAINT": "3",
    "FP_ENDCAP_CELL": "gf180mcu_fd_sc_mcu7t5v0__endcap",
    "PLACE_SITE_HEIGHT": "3.92",
    "DECAP_CELL": "gf180mcu_fd_sc_mcu7t5v0__fillcap_*",
    "CTS_ROOT_BUFFER": "gf180mcu_fd_sc_mcu7t5v0__clkbuf_16",
    "CTS_MAX_CAP": "0.5",
}


def test_pdk_migration_basics():
    from .pdk import migrate_old_config

    sky130A_migrated = migrate_old_config(RAW_SKY130A_IN)
    gf180mcuC_migrated = migrate_old_config(RAW_GF180MCUC_IN)

    for pdk_name, pdk_migrated in [
        ("sky130A", sky130A_migrated),
        ("gf180mcuC", gf180mcuC_migrated),
    ]:
        for bad_variable in [
            "SYNTH_DRIVING_CELL_PIN",
            "SYNTH_TIEHI_PORT",
            "SYNTH_TIELO_PORT",
            "SYNTH_MIN_BUF_PORT",
            "DIODE_CELL_PIN",
            "RCX_RULES",
            "RCX_RULES_MIN",
            "RCX_RULES_MAX",
            "TECH_LEF",
            "TECH_LEF_MIN",
            "TECH_LEF_MAX",
            "SYNTH_CAP_LOAD",
        ]:
            assert (
                bad_variable not in pdk_migrated
            ), f"OL2-incompatible variable {bad_variable} survived migration script for {pdk_name}"

        for cell_variable in [
            "SYNTH_DRIVING_CELL",
            "SYNTH_TIEHI_CELL",
            "SYNTH_BUFFER_CELL",
            "DIODE_CELL",
        ]:
            assert (
                cell_variable in pdk_migrated
            ), f"required variable {cell_variable} missing from migrated object for sky130"

            assert (
                "/" in pdk_migrated[cell_variable]
            ), f"required variable {cell_variable} has invalid value: {pdk_migrated[cell_variable]}"



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import pytest
import tkinter
from pyfakefs.fake_filesystem_unittest import Patcher


@pytest.fixture()
def _mock_fs():
    with Patcher() as patcher:
        patcher.fs.create_dir("/cwd")
        os.chdir("/cwd")
        yield


def test_escape():
    from .tcl import TclUtils

    assert (
        TclUtils.escape("ringo") == r"ringo"
    ), ".escape altered a non-dangerous string"
    assert TclUtils.escape("r ingo") == r'"r ingo"', ".escape neglected whitespace"
    assert (
        TclUtils.escape("[expr ringo]") == r'"\[expr ringo]"'
    ), ".escape neglected expression braces"
    assert (
        TclUtils.escape("[expr $ringo]") == r'"\[expr \$ringo]"'
    ), ".escape neglected dollar sign"
    assert (
        TclUtils.escape("[expr \\$ringo]") == r'"\[expr \\\$ringo]"'
    ), ".escape neglected backslash"


@pytest.mark.usefixtures("_mock_fs")
def test_join():
    from .tcl import TclUtils

    very_wild_list = [
        "{}{}{{}}{}{}{}{}{}{}{",
        "dollar\\ sign",
        "actual dollar $ign",
        '[exec {touch "need legal advice?"}]',
        '\\[exec {touch "better call saul"}]',
    ]

    very_wild_list_escaped = TclUtils.join(very_wild_list)
    interpreter = tkinter.Tcl()
    interpreter.eval(
        f"""
        set very_wild_variable [list {very_wild_list_escaped}]
        set length [llength $very_wild_variable]
        foreach i $very_wild_variable {{
            puts $i
        }}
        """
    )

    assert os.listdir("/cwd") == [], "Arbitrary code execution"
    assert interpreter.getvar("length") == 5, "Tcl list not escaped properly"


def test_eval_env():
    from .tcl import TclUtils

    env_backup = os.environ.copy()

    result = TclUtils._eval_env(
        {"DESIGN_DIR": "/cwd", "PDK": "sky130A", "TMPDIR": "/cwd"},
        """
        if { ![info exists ::env(STD_CELL_LIBRARY)] } {
            set ::env(STD_CELL_LIBRARY) "sky130_fd_sc_hd"
        }
        if { $::env(STD_CELL_LIBRARY) == "sky130_fd_sc_hd" } {
            set ::env(WHATEVER) 1
        } else {
            set ::env(WHATEVER) 0
        }
        """,
    )

    expected = {
        "DESIGN_DIR": "/cwd",
        "PDK": "sky130A",
        "TMPDIR": "/cwd",
        "WHATEVER": "1",
        "STD_CELL_LIBRARY": "sky130_fd_sc_hd",
    }

    assert result == expected, "conditional evaluation test failed"

    assert os.environ == env_backup, "_eval_env leaked environment changes"

    os.environ["STD_CELL_LIBRARY"] = "sky130_fd_sc_hd"

    TclUtils._eval_env(
        {},
        """
        set ::env(STD_CELL_LIBRARY) "sky130_fd_sc_hs"
        """,
    )

    assert (
        os.getenv("STD_CELL_LIBRARY") == "sky130_fd_sc_hd"
    ), "env_from_tcl unset a previously set environment variable"

    del os.environ["STD_CELL_LIBRARY"]

    assert result == expected



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import pytest
from decimal import Decimal
from dataclasses import dataclass
from pyfakefs.fake_filesystem_unittest import Patcher
from typing import Dict, List, Literal, Optional, Tuple, Type, Union


from .variable import Variable


@pytest.fixture()
def _mock_fs():
    with Patcher() as patcher:
        patcher.fs.create_dir("/cwd")
        os.chdir("/cwd")
        patcher.fs.create_file("/cwd/a")
        patcher.fs.create_file("/cwd/b")
        yield


def test_macro_validation():
    from . import Macro

    with pytest.raises(ValueError, match="at least one GDSII file"):
        Macro(gds=[], lef=["test"])

    with pytest.raises(ValueError, match="at least one LEF file"):
        Macro(gds=["test"], lef=[])


def test_is_optional():
    from .variable import is_optional

    assert is_optional(int) is False, "is_optional false positive"
    assert is_optional(Optional[int]) is True, "is_optional false negative"
    assert (
        is_optional(Optional[Union[int, dict]]) is True
    ), "is_optional composite false negative"
    assert (
        is_optional(Union[None, int, dict]) is True
    ), "is_optional flattened union false negative"


def test_some_of():
    from .variable import some_of

    assert some_of(int) == int, "some_of changed the type of a non-option type"
    assert (
        some_of(List[str]) == List[str]
    ), "some_of changed the type of a non-option type"
    assert (
        some_of(Optional[int]) == int
    ), "some_of failed to extract type from option type"
    assert (
        some_of(Optional[Union[Dict, List]]) == Union[Dict, List]
    ), "some of failed to properly handle optional union"

    assert (
        some_of(Union[Dict, List, None]) == Union[Dict, List]
    ), "some of failed to properly handle flattened optional union"


def test_variable_construction():
    from . import Variable

    variable = Variable(
        "EXAMPLE",
        Optional[int],
        description="My Description",
        deprecated_names=["OLD_EXAMPLE"],
    )

    assert variable.optional, ".optional property incorrectly set"
    assert variable.some == int, ".some property incorrectly set"

    variable_b = Variable(
        "EXAMPLE",
        Optional[int],
        description="My Other Description",
    )

    assert (
        variable == variable_b
    ), "Variable with different description or deprecated_name didn't match"

    variable_union = Variable(
        "UNION_VAR",
        Union[int, Dict[str, str]],
        description="x",
    )
    assert (
        variable_union.type == Union[int, Dict[str, str]]
    ), "Union magically switched types"


@pytest.fixture()
def variable():
    from .variable import Path, Variable

    return Variable(
        "EXAMPLE",
        Optional[List[Path]],
        description="x",
        deprecated_names=["OLD_EXAMPLE"],
    )


@pytest.mark.usefixtures("_mock_fs")
def test_compile(variable: Variable):
    from ..common import GenericDict

    valid_input = GenericDict({"EXAMPLE": ["/cwd/a", "/cwd/b"]})
    warning_list = []
    used_name, paths = variable.compile(
        valid_input,
        warning_list,
    )
    assert used_name == "EXAMPLE", "valid input returned incorrect used name"
    assert paths == [
        "/cwd/a",
        "/cwd/b",
    ], "valid input resolved paths incorrectly"
    assert len(warning_list) == 0, "valid input generated warning"


def test_compile_dataclass():
    from ..common import GenericDict

    @dataclass
    class MyClass:
        a: int
        b: str

    variable = Variable("MY_VARIABLE", MyClass, description="x")

    _, value = variable.compile(
        GenericDict({"MY_VARIABLE": {"a": 4, "b": "potato"}}), []
    )

    assert value == MyClass(a=4, b="potato"), "Failed to deserialize dataclass"


def test_compile_required():
    from ..common import GenericDict

    variable = Variable("MY_VARIABLE", int, description="x")

    with pytest.raises(
        ValueError,
        match=r"Required variable 'MY_VARIABLE' did not get a specified value",
    ):
        variable.compile(GenericDict({}), [])


@pytest.mark.usefixtures("_mock_fs")
def test_compile_deprecated(variable: Variable):
    from ..common import GenericDict

    deprecated_valid_input = GenericDict(
        {
            "OLD_EXAMPLE": [
                "/cwd/a",
                "/cwd/b",
            ]
        }
    )
    warning_list = []
    used_name, paths = variable.compile(
        deprecated_valid_input,
        warning_list,
    )
    assert (
        used_name == "OLD_EXAMPLE"
    ), "deprecated valid input returned incorrect used name"
    assert paths == [
        "/cwd/a",
        "/cwd/b",
    ], "deprecated valid input returned paths resolved incorrectly"
    assert len(warning_list) == 1, "use of deprecated names did not produce a warning"


@pytest.fixture()
def test_enum():
    from ..common import StringEnum

    return StringEnum("x", ["AValue", "AnotherValue"])


@pytest.fixture()
def variable_set(variable, test_enum):
    return [
        variable,
        Variable(
            "LIST_VAR",
            List[int],
            description="x",
        ),
        Variable(
            "TUPLE_2_VAR",
            Tuple[int, int],
            description="x",
        ),
        Variable(
            "TUPLE_3_VAR",
            Tuple[int, int, int],
            description="x",
        ),
        Variable(
            "DICT_VAR",
            Dict[str, str],
            description="x",
        ),
        Variable(
            "OTHER_DICT_VAR",
            Dict[str, str],
            description="x",
        ),
        Variable(
            "UNION_VAR",
            Union[int, Dict[str, str]],
            description="x",
        ),
        Variable(
            "LITERAL_VAR",
            Literal["yes"],
            description="x",
            default="yes",
        ),
        Variable(
            "BOOL_VAR",
            bool,
            description="x",
        ),
        Variable(
            "ENUM_VAR",
            test_enum,
            description="x",
        ),
        Variable(
            "NUMBER_VAR",
            Decimal,
            description="x",
        ),
    ]


@pytest.mark.usefixtures("_mock_fs")
def test_compile_invalid(variable_set: List[Variable]):
    from ..common import GenericDict

    invalid_input = GenericDict(
        {
            "EXAMPLE": [
                "/cwd/a",
                "/cwd/c",
            ],
            "LIST_VAR": {},
            "TUPLE_2_VAR": [1, {}],
            "TUPLE_3_VAR": [1, 4],
            "DICT_VAR": [],
            "OTHER_DICT_VAR": "bad tcl dictionary",
            "UNION_VAR": [],
            "LITERAL_VAR": "no",
            "BOOL_VAR": "No",
            "ENUM_VAR": "NotAValue",
            "NUMBER_VAR": "v",
        }
    )
    warning_list = []

    for variable in variable_set:
        print(f"* Testing {variable.name} ({variable.type})…")
        with pytest.raises(
            ValueError, match="(invalid)|(does not exist)|(is not 'yes')"
        ):
            variable.compile(
                invalid_input,
                warning_list,
                permissive_typing=True,
            )


@pytest.mark.usefixtures("_mock_fs")
def test_compile_permissive(variable_set: List[Variable], test_enum: Type):
    from ..common import GenericDict

    permissive_valid_input = GenericDict(
        {
            "EXAMPLE": "/cwd/a /cwd/b",
            "LIST_VAR": "4,5,6",
            "TUPLE_2_VAR": "1 2",
            "TUPLE_3_VAR": "1;2;3",
            "DICT_VAR": "key1 value1 key2 value2",
            "OTHER_DICT_VAR": "key1 value1 key2 value2",
            "UNION_VAR": "4",
            "BOOL_VAR": "0",
            "ENUM_VAR": "AValue",
            "NUMBER_VAR": "90123",
        }
    )

    final = {}

    for variable in variable_set:
        _, value = variable.compile(
            permissive_valid_input,
            [],
            permissive_typing=True,
        )
        final[variable.name] = value

    assert final == {
        "EXAMPLE": ["/cwd/a", "/cwd/b"],
        "LIST_VAR": [4, 5, 6],
        "TUPLE_2_VAR": ("1", "2"),
        "TUPLE_3_VAR": ("1", "2", "3"),
        "DICT_VAR": {"key1": "value1", "key2": "value2"},
        "OTHER_DICT_VAR": {"key1": "value1", "key2": "value2"},
        "UNION_VAR": 4,
        "BOOL_VAR": False,
        "ENUM_VAR": test_enum("AValue"),
        "NUMBER_VAR": Decimal("90123"),
        "LITERAL_VAR": "yes",
    }, "Permissive parsing mode returned an unexpected result"

    for variable in variable_set:
        if variable.name in ["LITERAL_VAR", "ENUM_VAR"]:
            continue
        print(f"* Testing {variable.name} ({variable.type})…")
        with pytest.raises(ValueError, match="Refusing"):
            variable.compile(
                permissive_valid_input,
                [],
            )



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest
import tkinter

from .tclstep import TclStep
from .test_step import mock_config  # noqa: F401
from ..state import State
from ..config.test_config import _mock_fs, mock_variables  # noqa: F401
from ..common import TclUtils


def test_tclstep_missing_get_script_path():
    class TclStepTest(TclStep):
        pass

    error = "Can't instantiate abstract class TclStepTest with abstract methods? get_script_path"  # Python 3.8 says "methods" even if there is one method
    with pytest.raises(TypeError, match=error):
        TclStepTest()


@pytest.mark.usefixtures("_mock_fs")
@mock_variables()
def test_tclstep_init(mock_config):  # noqa: F811
    from ..state.design_format import DesignFormat

    class TclStepTest(TclStep):
        inputs = []
        outputs = []
        id = "Test.TclStep"

        def get_script_path(self):
            return "/dummy_path"

    TclStepTest(config=mock_config, state_in=State({DesignFormat.NETLIST: "abc"}))


@pytest.mark.usefixtures("_mock_fs")
@mock_variables()
def test_tclstep_get_command(mock_config):  # noqa: F811
    from ..state.design_format import DesignFormat

    class TclStepTest(TclStep):
        inputs = []
        outputs = []
        id = "Test.TclStep"

        def get_script_path(self):
            return "/dummy_path"

    assert TclStepTest(
        config=mock_config, state_in=State({DesignFormat.NETLIST: "abc"})
    ).get_command() == ["tclsh", "/dummy_path"], "Wrong TclStep command"


def test_tcl_step_value_to_tcl():
    from enum import Enum

    numerical_list_value = [1.0, 2.1000, 3, 4.01]
    numerical_list_value_string = "1.0 2.1 3 4.01"
    assert (
        TclStep.value_to_tcl(numerical_list_value) == numerical_list_value_string
    ), "Wrong numerical list to tcl list conversion"

    alphanumerical_list_value = [
        1.0,
        2.1000,
        3,
        "abc",
        "abc f",
        "{abc}",
        "[abc]",
        "\\abc",
    ]
    alphanumerical_list_value_string = (
        '1.0 2.1 3 abc "abc f" "{abc}" "\\[abc]" "\\\\abc"'
    )
    assert (
        TclStep.value_to_tcl(alphanumerical_list_value)
        == alphanumerical_list_value_string
    ), "Wrong alpahnumerical list to tcl list conversion"

    assert TclStep.value_to_tcl(True) == "1", "Wrong bool to tcl conversion"
    assert TclStep.value_to_tcl(1) == "1", "Wrong integer to tcl conversion"
    assert TclStep.value_to_tcl(-1) == "-1", "Wrong negative integer to tcl conversion"
    assert TclStep.value_to_tcl(-1.01) == "-1.01", "Wrong decimal to tcl conversion"
    assert TclStep.value_to_tcl("Hello") == "Hello", "Wrong string to tcl conversion"

    assert TclStep.value_to_tcl({}) == ""
    assert TclStep.value_to_tcl([]) == ""
    assert TclStep.value_to_tcl("") == ""
    assert TclStep.value_to_tcl(" ") == " "

    class Color(Enum):
        RED = 1

    assert TclStep.value_to_tcl(Color.RED) == "RED"

    dict_value = {"abc": 1}
    dict_value_string = "abc 1"
    assert (
        TclStep.value_to_tcl(dict_value) == dict_value_string
    ), "Wrong dict to tcl conversion"
    dict_value = {
        "abc": 1,
        "cde": "123",
        "f": "[hello]",
        "h": [1, 2, 3],
        "[g]": alphanumerical_list_value,
    }

    interpreter = tkinter.Tcl()
    interpreter.eval(
        f"""
        set test_dict {TclUtils.escape(TclStep.value_to_tcl(dict_value))}
        set counter 0
        dict for {{k v}} $test_dict {{
            set "key$counter" $k
            set "value$counter" $v
            incr counter
        }}
        """
    )

    expected_dict_value = {
        "abc": "1",
        "cde": "123",
        "f": "[hello]",
        "h": "1 2 3",
        "[g]": alphanumerical_list_value_string,
    }

    out = {}
    counter = 0
    while True:
        key_var = f"key{counter}"
        value_var = f"value{counter}"

        try:
            key = interpreter.getvar(key_var)
        except tkinter.TclError:
            break
        value = interpreter.getvar(value_var)
        out[key] = value
        counter += 1

    assert out == expected_dict_value, "failed to convert dictionary"


@pytest.mark.usefixtures("_mock_fs")
@mock_variables()
def test_env(mock_config):  # noqa: F811
    from ..state.design_format import DesignFormat

    script_path = "/dummy_path"
    state_in = State({DesignFormat.NETLIST: "abc"})

    class TclStepTest(TclStep):
        inputs = [DesignFormat.NETLIST]
        outputs = [DesignFormat.NETLIST]
        id = "Test.TclStep"
        step_dir = "/dummy_step_dir"

        def get_script_path(self):
            return script_path

    step = TclStepTest(config=mock_config, state_in=state_in)
    env = step.prepare_env({}, state_in)
    assert (
        env["STEP_DIR"] == TclStepTest.step_dir
    ), "Wrong prepared env. Bad STEP_DIR value"
    assert env["CURRENT_NETLIST"] == "abc", "Wrong prepared env. Bad CURRENT_ input"
    assert "SAVE_NETLIST" in env, "Wrong prepared env. SAVE_NETLIST missing"
    for var in mock_config:
        if mock_config[var] is not None:
            assert var in env, "Wrong prepared env. Missing config variable"
            assert env[var] == TclStep.value_to_tcl(
                mock_config[var]
            ), "Wrong prepared env. Mismatching configuration variable"
